import distutils.spawn
import logging
import os
import re
import shutil
import subprocess
import tarfile
import uuid
from abc import abstractmethod
from io import BytesIO
from pathlib import Path, PurePath
from typing import TYPE_CHECKING, Generic, Iterable, List, Mapping, Optional, Pattern

import attr
from semantic_version import SimpleSpec, Version

from bento.error import DockerFailureException, ToolRunException
from bento.tool.tool import R, Tool
from bento.util import Memo


def __extract_use_binary() -> Optional[bool]:
    e = os.environ.get("BENTO_USE_BINARY")
    return e == "1" if e is not None else None


DOCKER_INSTALLED = Memo[bool](lambda: shutil.which("docker") is not None)
DEFAULT_VERSION_PATTERN = re.compile(r"(.*)")
USE_BINARY = __extract_use_binary()

if TYPE_CHECKING:
    import docker.client
    from docker.models.containers import Container


def get_docker_client() -> "docker.client.DockerClient":
    """Checks that docker client is reachable"""
    # import inside def for performance
    import docker

    try:
        client = docker.from_env()
        client.info()
        return client
    except Exception as e:
        logging.debug(e)
        raise DockerFailureException()


def copy_into_container(
    paths: Mapping[Path, str], container: "Container", destination_path: PurePath
) -> None:
    """Copy local ``paths`` to ``destination_path`` within ``container``.

    The ``container`` is assumed to be running.
    If ``destination_path`` does not exist, it will be created.
    """
    tar_buffer = BytesIO()
    with tarfile.open(mode="w", fileobj=tar_buffer) as archive:
        for p, loc in paths.items():
            archive.add(str(p), arcname=loc)
    tar_buffer.seek(0)
    tar_bytes = tar_buffer.read()

    logging.info(f"sending {len(tar_bytes)} bytes to {container}")
    container.put_archive(str(destination_path), tar_bytes)


def has_allowed_binary_version(
    binary: str, query: List[str], match: Pattern, spec: SimpleSpec
) -> bool:
    """
    Queries the version of a locally installed binary to determine if Docker can be skipped

    :param binary: The name of the binary executable
    :param query: Flags to pass to the executable to query the version
    :param match: Pattern used to extract the binary version from the query result
    :param spec: Allowed range of versions for skipping Docker
    :return: If Docker can be skipped
    """
    if USE_BINARY is False:
        logging.debug(f"Using Docker for {binary} due to passed BENTO_USE_BINARY flag.")
        return False
    executable = distutils.spawn.find_executable(binary)
    if not executable:
        logging.debug(f"Binary for '{binary}' not in path")
        if USE_BINARY is True:
            raise ToolRunException(f"Could not find '{binary}' in path.")
        return False
    query_output = subprocess.run(
        [executable] + query, stdout=subprocess.PIPE, encoding="utf-8"
    ).stdout.rstrip()
    version_match = match.search(query_output)
    logging.debug(f"'{binary}' version string is '{version_match}' in\n{query_output}")
    if not version_match or len(version_match.groups()) < 1:
        if USE_BINARY is True:
            cmd = " ".join([executable] + query)
            raise ToolRunException(
                f"Could not read version from '{cmd}' (got '{query_output}')."
            )
        return False
    if Version(version_match[1]) not in spec:  # type: ignore
        if USE_BINARY is True:
            raise ToolRunException(
                f"Version for '{binary}' is '{version_match[1]}', expected a version in the range '{spec}'."
            )
        return False
    return True


@attr.s
class DockerTool(Generic[R], Tool[R]):
    UUID = str(uuid.uuid4())[:8]

    use_binary = attr.ib(type=bool)

    @use_binary.default
    def default_use_binary(self) -> bool:
        if self.binary_executable is None:
            return False

        return has_allowed_binary_version(
            self.binary_executable,
            self.binary_version_query,
            self.binary_version_pattern,
            self.allowed_binary_versions,
        )

    @property
    @abstractmethod
    def docker_image(self) -> str:
        """
        Returns the name of the docker image
        """
        pass

    @property
    @abstractmethod
    def docker_command(self) -> List[str]:
        """
        Returns the command to run in the docker image
        """
        pass

    @property
    def use_remote_docker(self) -> bool:
        """Return whether the Docker daemon is remote.

        We need to know because volume mounting doesn't work with remote daemons.
        """
        return (
            os.getenv("BENTO_REMOTE_DOCKER", "0") == "1"
            or os.getenv("R2C_USE_REMOTE_DOCKER", "0") == "1"
        )

    @property
    def binary_executable(self) -> Optional[str]:
        """
        If present, this executable will be used to run instead of Docker
        :return:  The name of the binary, or None to always use Docker
        """
        return None

    @property
    def binary_version_query(self) -> List[str]:
        """
        Flags passed to binary to read version
        """
        return ["--version"]

    @property
    def binary_version_pattern(self) -> Pattern:
        """
        Pattern used to extract version from binary version output; version should appear in first capturing group
        """
        return DEFAULT_VERSION_PATTERN

    @property
    def allowed_binary_versions(self) -> SimpleSpec:
        """
        Allowed range of binary versions for which Docker should be skipped
        """
        return SimpleSpec(">=0.0.0")

    @property
    def binary_command(self) -> List[str]:
        """
        Binary command to use when skipping Docker
        """
        be = self.binary_executable
        return [be] if be else []

    @property
    def additional_file_targets(self) -> Mapping[Path, str]:
        """
        Additional files, beyond the files to check, that should be copied to the Docker container

        For example, this might include configuration files.

        Format is mapping of local path objects to remote path strings.
        """
        return {}

    @property
    @abstractmethod
    def remote_code_path(self) -> str:
        """
        Where to locate remote code within the docker container
        """
        pass

    @property
    def container_name(self) -> str:
        """
        The name of the docker container

        This is used to separate the docker instances used by each instance of Bento and
        each instance of this class within Bento.
        """
        # UUID separates instances of Bento
        # Tool ID separates instances of the tool
        return f"bento-daemon-{self.UUID}-{self.tool_id()}"

    @property
    def local_volume_mapping(self) -> Mapping[str, Mapping[str, str]]:
        """The volumes to bind when Docker is running locally"""
        return {str(self.base_path): {"bind": self.remote_code_path, "mode": "ro"}}

    def is_allowed_returncode(self, returncode: int) -> bool:
        """Returns true iff the Docker container's return code indicates no error"""
        return returncode == 0

    def assemble_full_command(self, targets: Iterable[str]) -> List[str]:
        """Creates the full command to send to the docker container from file targets"""
        return self.docker_command + list(targets)

    def _prepull_image(self) -> None:
        """
        Pulls the docker image from Docker hub
        """
        client = get_docker_client()

        if not any(i for i in client.images.list() if self.docker_image in i.tags):
            client.images.pull(self.docker_image)
            logging.info(f"Pre-pulled {self.tool_id()} image")

    def _create_container(self, targets: Iterable[str]) -> "Container":
        """
        Creates and returns the Docker container
        """

        client = get_docker_client()

        our_containers = client.containers.list(
            filters={"name": self.container_name, "status": "running"}
        )
        if our_containers:
            logging.info(f"using existing {self.tool_id()} container")
            return our_containers[0]

        vols = {} if self.use_remote_docker else self.local_volume_mapping
        full_command = self.assemble_full_command(targets)

        logging.info(f"starting new {self.tool_id()} container")

        container: "Container" = client.containers.create(
            self.docker_image,
            command=full_command,
            auto_remove=True,
            name=self.container_name,
            volumes=vols,
            detach=True,
            working_dir=self.remote_code_path,
        )
        logging.info(f"started container: {container!r}")

        return container

    def _setup_remote_docker(
        self, container: "Container", expanded: Mapping[Path, str]
    ) -> None:
        """Copies files into the remote Docker container"""
        copy_into_container(expanded, container, PurePath(self.remote_code_path))
        copy_into_container(
            self.additional_file_targets, container, PurePath(self.remote_code_path)
        )

    def _run_container(self, files: Iterable[str]) -> subprocess.CompletedProcess:
        logging.info(f"Running {self.tool_id()} using Docker image")
        targets: Iterable[Path] = [Path(p) for p in files]
        is_remote = self.use_remote_docker
        expanded = {t: str(t.relative_to(self.base_path)) for t in targets}

        container = self._create_container(expanded.values())

        if is_remote:
            self._setup_remote_docker(container, expanded)

        # Python docker does not allow -a, so use subprocess.run
        return subprocess.run(
            ["docker", "start", "-a", str(container.id)],
            encoding="utf-8",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    def _run_binary(self, files: Iterable[str]) -> subprocess.CompletedProcess:
        logging.info(f"Running {self.tool_id()} using installed binary")
        command = self.binary_command + self.assemble_full_command(files)
        return subprocess.run(
            command, encoding="utf-8", stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

    def run_command(self, files: Iterable[str]) -> subprocess.CompletedProcess:
        """
        Run this tool, either in a local binary or Docker container
        """
        result = (
            self._run_binary(files) if self.use_binary else self._run_container(files)
        )

        logging.info(
            f"{self.tool_id()}: Returned code {result.returncode} with stdout[:4000]:\n"
            f"{result.stdout[:4000]}\nstderr[:4000]\n{result.stderr[:4000]}"
        )

        if not self.is_allowed_returncode(result.returncode):
            raise subprocess.CalledProcessError(
                cmd=self.docker_command,
                returncode=result.returncode,
                output=result.stdout,
                stderr=result.stderr,
            )

        return result

    def matches_project(self, files: Iterable[Path]) -> bool:
        return (
            DOCKER_INSTALLED.value or not self.use_binary
        ) and self.project_has_file_paths(files)

    def setup(self) -> None:
        if not self.use_binary:
            self._prepull_image()



import json
import re
from pathlib import Path
from typing import Iterable, List, Optional, Pattern, Type

from semantic_version import SimpleSpec

from bento.parser import Parser
from bento.tool import output, runner
from bento.violation import Violation


"""
Input example:

[
  {
    "message": "flask apps using 'flask-wtf' require including a csrf token in the html form. this check detects missing csrf protection in html forms in jinja templates.",
    "physical_line": "        <form method=\"post\">",
    "code": "jinjalint-form-missing-csrf-protection",
    "file_path": "jinja-template.html",
    "line": 7,
    "column": 8
  },
  {
    "message": "pages opened with 'target=\"_blank\"' allow the new page to access the original's 'window.opener'. this can have security and performance implications. include 'rel=\"noopener\"' to prevent this.",
    "physical_line": "        <a href=\"https://example.com\" target=\"_blank\">test anchor</a>",
    "code": "jinjalint-anchor-missing-noopener",
    "file_path": "jinja-template.html",
    "line": 11,
    "column": 8
  },
  {
    "message": "pages opened with 'target=\"_blank\"' allow the new page to access the original's referrer. this can have privacy implications. include 'rel=\"noreferrer\"' to prevent this.",
    "physical_line": "        <a href=\"https://example.com\" target=\"_blank\">test anchor</a>",
    "code": "jinjalint-anchor-missing-noreferrer",
    "file_path": "jinja-template.html",
    "line": 11,
    "column": 8
  }
]
"""


class JinjalintParser(Parser[str]):
    CHECK_PREFIX = "jinjalint-"
    CHECK_PREFIX_LEN = len(CHECK_PREFIX)
    SEVERITY = {"LOW": 0, "MEDIUM": 1, "HIGH": 2}

    @staticmethod
    def _get_link(code: str) -> Optional[str]:
        if "recursion-error" in code or "parse-error" in code:
            # These errors do not have documentation
            return None

        code_suffix = code.replace(JinjalintParser.CHECK_PREFIX, "")
        return f"https://bento.dev/checks/jinja/{code_suffix}/"

    def parse(self, tool_output: str) -> List[Violation]:
        results = json.loads(tool_output)

        violations = [
            Violation(
                check_id=r["code"].replace(self.CHECK_PREFIX, ""),
                tool_id=JinjalintTool.TOOL_ID,
                path=self.trim_base(r["file_path"]),
                severity=JinjalintParser.SEVERITY["MEDIUM"],
                line=r["line"],
                column=r["column"],
                message=r["message"],
                syntactic_context=r.get("physical_line", ""),
                link=self._get_link(r["code"]),
            )
            for r in results
        ]

        return violations


class JinjalintTool(runner.Python, output.Str):
    TOOL_ID = "r2c.jinja"
    VENV_DIR = "jinjalint"
    PROJECT_NAME = "Python"
    JINJA_FILE_NAME_FILTER = re.compile(
        r".*\.(html|jinja|jinja2|j2|twig)$"
    )  # Jinjalint's default extensions
    PACKAGES = {"r2c-jinjalint": SimpleSpec("~=0.7.1")}

    @property
    def shebang_pattern(self) -> Optional[Pattern]:
        return None

    @property
    def parser_type(self) -> Type[Parser]:
        return JinjalintParser

    @classmethod
    def tool_id(self) -> str:
        return JinjalintTool.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Finds common security issues in Jinja templates"

    @classmethod
    def venv_subdir_name(cls) -> str:
        return JinjalintTool.VENV_DIR

    @property
    def project_name(self) -> str:
        return JinjalintTool.PROJECT_NAME

    @property
    def file_name_filter(self) -> Pattern:
        return JinjalintTool.JINJA_FILE_NAME_FILTER

    def matches_project(self, files: Iterable[Path]) -> bool:
        has_jinja = any((self.JINJA_FILE_NAME_FILTER.match(p.name) for p in files))
        has_python = any((self.PYTHON_FILE_PATTERN.match(p.name) for p in files))
        return has_jinja and has_python

    def run(self, paths: Iterable[str]) -> str:
        launchpoint: str = str(self.venv_dir() / "bin" / "jinjalint")
        exclude_rules = [
            "--exclude",
            "jinjalint-space-only-indent",
            "--exclude",
            "jinjalint-misaligned-indentation",
        ]
        cmd = ["python", launchpoint, "--json"] + exclude_rules + list(paths)
        return self.venv_exec(cmd, check_output=False)



import logging
import os
from typing import Any, Callable, List, Set, cast

import click
import yaml

import bento.extra
from bento.context import Context
from bento.error import InvalidToolException
from bento.util import AutocompleteSuggestions


def update_tool_run(context: Context, tool: str, run: bool) -> None:
    """
    Sets run field of tool to RUN. Default to no ignore if tool not in config
    """

    config = context.config
    tool_config = config["tools"]
    all_tools = ", ".join(f"'{k}'" for k in tool_config.keys())
    # Check that TOOL is valid
    if tool not in {t.tool_id() for t in bento.extra.TOOLS}:
        raise InvalidToolException(tool, all_tools)

    if tool not in tool_config:
        # Read default ignore from default config file
        with (
            open(os.path.join(os.path.dirname(__file__), "configs/default.yml"))
        ) as template:
            yml = yaml.safe_load(template)

        default_ignore: List[str] = []
        if tool in yml["tools"]:
            default_ignore = yml["tools"][tool]["ignore"]

        tool_config[tool] = {"ignore": default_ignore}

    tool_config[tool]["run"] = run
    context.config = config


def update_ignores(
    context: Context, tool: str, update_func: Callable[[Set[str]], None]
) -> None:
    config = context.config
    tool_config = config["tools"]
    if tool not in tool_config:
        all_tools = ", ".join(f"'{k}'" for k in tool_config.keys())
        raise InvalidToolException(tool, all_tools)

    ignores = set(tool_config[tool].get("ignore", []))
    update_func(ignores)

    tool_config[tool]["ignore"] = sorted(list(ignores))

    context.config = config


def get_valid_tools(
    ctx: Any, args: List[str], incomplete: str
) -> AutocompleteSuggestions:
    """
        Return list of tools supported by bento

        Hides experimental tools
    """
    # context is not yet initialized, so just do it now
    try:
        tool_list = sorted(
            [
                (t.tool_id(), t.tool_desc())
                for t in bento.extra.TOOLS
                if t not in bento.extra.EXPERIMENTAL_TOOLS
            ],
            key=(lambda td: td[0]),
        )
        results = list(filter(lambda s: s[0].startswith(incomplete), tool_list))
        return cast(AutocompleteSuggestions, results)
    except Exception as ex:
        logging.warning(ex)
        return []


def get_tool_help(summary: str) -> str:
    tools = get_valid_tools(None, [], "")
    max_width = max(len(t) for t, _ in tools)
    lines = (f"{t:<{max_width}s} -- {d}" for t, d in tools)
    result = "\n    ".join(lines)
    return f"""
  {summary}

  Possible tools are:

    {result}"""


class ToolCommand(click.Command):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.help_summary = kwargs["help_summary"]
        del kwargs["help_summary"]
        super().__init__(*args, **kwargs)

    def format_help_text(self, _: Any, formatter: click.HelpFormatter) -> None:
        formatter.write(get_tool_help(self.help_summary))
        formatter.write("\n")


def get_disabled_checks(
    ctx: Any, args: List[str], incomplete: str
) -> AutocompleteSuggestions:
    # context is not yet initialized, so just do it now
    try:
        context = Context()
        results = list(
            filter(
                lambda s: s.startswith(incomplete),
                context.config["tools"][args[-1]]["ignore"],
            )
        )
        return cast(AutocompleteSuggestions, results)
    except Exception:
        return []



import json
from typing import Any, Dict, Iterable, List, Mapping, Type

from semantic_version import SimpleSpec

from bento.parser import Parser
from bento.tool import output, runner
from bento.violation import Violation

# Input example:
# {
#     "./foo.py": [
#         {
#             "code": "E124",
#             "filename": "./foo.py",
#             "line_number": 2,
#             "column_number": 9,
#             "text": "closing bracket does not match visual indentation",
#             "physical_line": "        )\n"
#         }
#     ]
# }
#

# Only these prefixes will be inspected
RULE_PREFIXES = "B,C90,E113,E74,E9,EXE,F,T100,W6"

FLAKE8_TO_BENTO = {
    "B001": "bare-except-bugbear",
    "B002": "unsupported-unary-increment",
    "B003": "assignment-to-environ",
    "B004": "unreliable-hasattr-call",
    "B005": "no-multicharacter-strip",
    "B006": "no-mutable-default-args",
    "B007": "unused-loop-variable",
    "B008": "no-call-in-default-args",
    "B009": "no-getattr",
    "B010": "no-setattr",
    "B011": "no-assert-false",
    "B012": "escaped-finally",
    "B301": "no-iter-methods",
    "B302": "no-view-methods",
    "B303": "no-metaclass",
    "B304": "no-maxint",
    "B305": "no-next-method",
    "B306": "no-exception-message-method",
    "C90": "too-complex",
    "E113": "indentation-error",
    "E701": "multiple-statements-colon",
    "E702": "multiple-statements-semicolon",
    "E703": "ending-semicolon",
    "E704": "multiple-statements-def",
    "E711": "none-comparison",
    "E712": "true-comparison",
    "E713": "membership-test",
    "E714": "identity-test",
    "E721": "type-comparison",
    "E722": "bare-except",
    "E731": "no-assign-lambda",
    "E741": "ambiguous-variable-name",
    "E742": "ambiguous-class-name",
    "E743": "ambiguous-function-name",
    "E901": "syntax-error",
    "E902": "io-error",
    "E999": "parse-error",
    "EXE001": "not-executable",
    "EXE002": "missing-shebang",
    "EXE003": "not-python",
    "EXE004": "whitespace-before-shebang",
    "EXE005": "lines-before-shebang",
    "F401": "unused-module",
    "F402": "shadowed-import",
    "F403": "wildcard-import",
    "F404": "future-import-order",
    "F405": "undefined-module-name",
    "F406": "bad-wildcard-import",
    "F601": "dict-key-repeated",
    "F602": "dict-key-variable-repeated",
    "F621": "too-many-expressions-with-star",
    "F622": "more-than-one-star-in-assignment",
    "F631": "assert-tuple-always-true",
    "F632": "use-eqeq-for-literals",
    "F633": "invalid-arrows-with-print",
    "F811": "redefined-unused-name",
    "F812": "redefined-name",
    "F821": "undefined-name",
    "F822": "undefined-name-in-all",
    "F823": "unassigned-variable",
    "F831": "duplicated-argument",
    "F841": "unused-variable",
    "T100": "debugger",
    "W601": "deprecated-has-key",
    "W602": "deprecated-raise",
    "W603": "deprecated-inequality",
    "W604": "deprecated-backticks",
    "W605": "invalid-escape",
    "W606": "bad-async-or-await",
}


class Flake8Parser(Parser[str]):
    @classmethod
    def to_bento(cls) -> Mapping[str, str]:
        return FLAKE8_TO_BENTO

    @staticmethod
    def id_to_link(check_id: str) -> str:
        if check_id == "E999":
            link = ""
        elif check_id.startswith("B"):
            link = "https://github.com/PyCQA/flake8-bugbear/blob/master/README.rst#list-of-warnings"
        elif check_id == "T100":
            link = "https://github.com/JBKahn/flake8-debugger/blob/master/README.md"
        elif check_id.startswith("EXE"):
            link = "https://github.com/xuhdev/flake8-executable/blob/master/README.md#flake8-executable"
        elif check_id.startswith("F6"):
            link = "https://flake8.pycqa.org/en/latest/user/error-codes.html"
        elif check_id in ("E722", "E306", "E117", "E113", "W606"):
            link = "https://pycodestyle.readthedocs.io/en/latest/intro.html#error-codes"
        else:
            link = f"https://lintlyci.github.io/Flake8Rules/rules/{check_id}.html"
        return link

    @classmethod
    def id_to_name(cls, check_id: str) -> str:
        return cls.to_bento().get(check_id, check_id)

    @staticmethod
    def tool() -> Type[output.Str]:
        return Flake8Tool

    def to_violation(self, result: Dict[str, Any]) -> Violation:
        source = (result["physical_line"] or "").rstrip()  # Remove trailing whitespace
        path = self.trim_base(result["filename"])

        check_id = result["code"]

        return Violation(
            tool_id=self.tool().tool_id(),
            check_id=self.id_to_name(check_id),
            path=path,
            line=result["line_number"],
            column=result["column_number"],
            message=result["text"],
            severity=2,
            syntactic_context=source,
            link=self.id_to_link(check_id),
        )

    def parse(self, tool_output: str) -> List[Violation]:
        results: Dict[str, List[Dict[str, Any]]] = json.loads(tool_output)
        return [self.to_violation(v) for r in results.values() for v in r]


class Flake8Tool(runner.Python, output.Str):
    TOOL_ID = "flake8"  # to-do: versioning?
    VENV_DIR = "flake8"
    PROJECT_NAME = "Python"
    PACKAGES = {
        "flake8": SimpleSpec("~=3.7.0"),
        "flake8-json": SimpleSpec("~=19.8.0"),
        "flake8-bugbear": SimpleSpec("~=20.1.4"),
        "flake8-debugger": SimpleSpec("~=3.2.0"),
    }

    @property
    def parser_type(self) -> Type[Parser]:
        return Flake8Parser

    @classmethod
    def tool_id(cls) -> str:
        return cls.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Finds common bugs in Python code"

    @property
    def project_name(self) -> str:
        return self.PROJECT_NAME

    @classmethod
    def venv_subdir_name(self) -> str:
        return self.VENV_DIR

    def select_clause(self) -> str:
        """Returns a --select argument to identify which checks flake8 should run"""
        return f"--select={RULE_PREFIXES}"

    def run(self, paths: Iterable[str]) -> str:
        cmd = [
            "python",
            str(self.venv_dir() / "bin" / "flake8"),
            self.select_clause(),
            "--format=json",
            "--isolated",
            *paths,
        ]
        return self.venv_exec(cmd, check_output=False)



import json
import logging
import os
import re
import subprocess
import sys
import venv
from abc import abstractmethod
from pathlib import Path
from time import time
from typing import Dict, Generic, Iterable, List, Optional, Pattern

from semantic_version import SimpleSpec, Version

import bento.constants as constants
from bento.tool.tool import R, Tool


class PythonTool(Generic[R], Tool[R]):
    # On most environments, just "pip" will point to the wrong Python installation
    # Fix by using the virtual environment's python3
    # we hard code to python3
    PIP_CMD = ["python3", "-m", "pip"]
    PACKAGES: Dict[str, SimpleSpec] = {}
    PYTHON_FILE_PATTERN = re.compile(r".*\.py$")
    SHEBANG_PATTERN = re.compile(r"^#!.*python")

    @property
    def shebang_pattern(self) -> Optional[Pattern]:
        return self.SHEBANG_PATTERN

    def matches_project(self, files: Iterable[Path]) -> bool:
        return self.project_has_file_paths(files)

    @property
    def file_name_filter(self) -> Pattern:
        return self.PYTHON_FILE_PATTERN

    @classmethod
    @abstractmethod
    def venv_subdir_name(cls) -> str:
        pass

    @classmethod
    def required_packages(cls) -> Dict[str, SimpleSpec]:
        return cls.PACKAGES

    @classmethod
    def venv_dir(cls) -> Path:
        return constants.VENV_PATH / cls.venv_subdir_name()

    def venv_create(self) -> None:
        """
        Creates a virtual environment for this tool
        """
        if not self.venv_dir().exists():
            logging.info(f"Creating virtual environment for {self.tool_id()}")
            # If we are already in a virtual environment, venv.create() will fail to install pip,
            # but we probably have virtualenv in the path, so try that first.
            try:
                # Don't litter stdout with virtualenv spam
                # Create virtualenv using same version as currently running executable
                subprocess.run(
                    ["virtualenv", f"--python={sys.executable}", self.venv_dir()],
                    stdout=subprocess.DEVNULL,
                    check=True,
                )
            except Exception:
                venv.create(str(self.venv_dir()), with_pip=True)

    def venv_exec(self, cmd: List[str], check_output: bool = True) -> str:
        """
        Executes tool set-up or check within its virtual environment
        """
        logging.debug(f"{self.tool_id()}: Running '{cmd}'")
        before = time()
        env = dict(os.environ)
        env["VIRTUAL_ENV"] = str(self.venv_dir())
        env["PATH"] = f"{self.venv_dir()}:{self.venv_dir()}/bin:" + env["PATH"]
        if "PYTHONHOME" in env:
            del env["PYTHONHOME"]
        v = subprocess.Popen(
            cmd,
            cwd=str(self.base_path),
            encoding="utf8",
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = v.communicate()
        after = time()
        logging.debug(f"{self.tool_id()}: Command completed in {after - before:2f} s")
        logging.debug(f"{self.tool_id()}: stderr[:4000]:\n" + stderr[0:4000])
        logging.debug(f"{self.tool_id()}: stdout[:4000]:\n" + stdout[0:4000])
        if check_output and v.returncode != 0:
            raise subprocess.CalledProcessError(
                v.returncode, cmd, output=stdout, stderr=stderr
            )
        return stdout

    def _packages_installed(self) -> Dict[str, SimpleSpec]:
        """
        Checks whether the given packages are installed.

        The value for each package is the version specification.
        """
        installed: Dict[str, Version] = {}
        for package in json.loads(
            self.venv_exec([*PythonTool.PIP_CMD, "list", "--format", "json"])
        ):
            try:
                installed[package["name"]] = Version(package["version"])
            except ValueError:
                # skip it
                pass

        to_install: Dict[str, SimpleSpec] = {}
        for name, spec in self.required_packages().items():
            if name not in installed or not spec.match(installed[name]):
                to_install[name] = spec
        return to_install

    def setup(self) -> None:
        self.venv_create()
        to_install = self._packages_installed()
        if not to_install:
            return

        install_list = [f"{p}{s.expression}" for p, s in to_install.items()]
        logging.info(f"Installing Python packages: {', '.join(install_list)}")
        self.venv_exec(
            [*PythonTool.PIP_CMD, "install", "-q", *install_list], check_output=True
        )



import configparser
from pathlib import Path
from typing import TYPE_CHECKING, Optional

# XXX: This is hacky. This should maybe use the Context object or something to
# determine the base directory.

if TYPE_CHECKING:
    # Only import when type checking to avoid loading module when unecessary
    import git  # noqa


def repo(path: Optional[Path] = None) -> Optional["git.Repo"]:
    # import inside def for performance
    import git.exc

    try:
        r = git.Repo(str(path or Path.cwd()), search_parent_directories=True)
        return r
    except git.exc.InvalidGitRepositoryError:
        return None


# N.B. See https://stackoverflow.com/a/42613047
def user_email(path: Optional[Path] = None) -> Optional[str]:
    r = repo(path)
    if r is None:
        return None
    try:
        return r.config_reader().get_value("user", "email").strip("\"'")
    except configparser.NoSectionError:
        return None
    except configparser.NoOptionError:
        return None


def global_ignore_path(path: Optional[Path] = None) -> Optional[Path]:
    r = repo(path)
    if r is None:
        return None

    try:
        config_value = r.config_reader("global").get_value("core", "excludesfile")
        return Path(config_value.strip()).expanduser()
    except configparser.NoSectionError:
        return None
    except configparser.NoOptionError:
        return None


def url(path: Optional[Path] = None) -> Optional[str]:
    """Get remote.origin.url for git dir at dirPath"""
    r = repo(path)
    if r and r.remotes:
        if any(rr.name == "origin" for rr in r.remotes):
            return r.remotes.origin.url
        else:
            return r.remotes[0].url
    else:
        return None


def commit(path: Optional[Path] = None) -> Optional[str]:
    """Get head commit for git dir at dirPath"""
    r = repo()
    if r is None:
        return None
    try:
        return str(r.head.commit)
    except ValueError:
        # catch case where local git repo without remote master
        return None



import os
from pathlib import Path

from click.termui import style

# Paths and file names

GLOBAL_RESOURCE_PATH = Path(os.path.expanduser("~/.bento"))
GLOBAL_CONFIG_PATH = GLOBAL_RESOURCE_PATH / "config.yml"
DEFAULT_LOG_PATH = GLOBAL_RESOURCE_PATH / "last.log"
VENV_PATH = GLOBAL_RESOURCE_PATH / "venv"
DEFAULT_GLOBAL_GIT_IGNORE_PATH = Path(os.path.expanduser("~/.config/git/ignore"))
GLOBAL_VERSION_CACHE_PATH = GLOBAL_RESOURCE_PATH / "version"

RESOURCE_PATH = Path(".bento")
CACHE_PATH = Path("cache")

ARCHIVE_FILE_NAME = "archive.json"
CONFIG_FILE_NAME = "config.yml"
IGNORE_FILE_NAME = ".bentoignore"
GREP_CONFIG_FILE_NAME = "grep-config.yml"
GH_ACTIONS_FILE_NAME = ".github/workflows/bento.yml"

# Registration data

TERMS_OF_SERVICE_KEY = "terms_of_service"
TERMS_OF_SERVICE_VERSION = "0.3.0"
GLOBAL_GIT_IGNORE_OPT_OUT = "opt_out_global_git_ignore"

# Identifiers

BENTO_TEMPLATE_HASH = "3a04e0f0cd9243d20b1e33da7ac13115"

BENTO_EMAIL_VAR = "BENTO_EMAIL"
BENTO_TEST_VAR = "BENTO_TEST"
QA_TEST_EMAIL_ADDRESS = "test@returntocorp.com"
SUPPORT_EMAIL_ADDRESS = "support@r2c.dev"

# Metrics constants

ARGS_TO_EXCLUDE_FROM_METRICS = {"check": {"paths"}}

# Config items

AUTORUN = "autorun"
AUTORUN_BLOCK = "block"

### messages ###

UPGRADE_WARNING_OUTPUT = f"""
╭─────────────────────────────────────────────╮
│  🎉 A new version of Bento is available 🎉  │
│  Try it out by running:                     │
│                                             │
│       {style("pip3 install --upgrade bento-cli", fg="blue")}      │
│                                             │
╰─────────────────────────────────────────────╯
"""



from bento.tool.runner.docker import DockerTool as Docker  # noqa
from bento.tool.runner.js_tool import JsTool as Node  # noqa
from bento.tool.runner.python_tool import PythonTool as Python  # noqa

"""
Classes that define what runs a tool:

runner.Docker - Runs inside a Docker container
runner.Node - Installs with npm / yarn, runs with node
runner.Python - Installs with venv and pip3, runs with python3
"""



import collections
import itertools
import logging
from abc import ABC, abstractmethod
from typing import Any, Collection, Iterator, Mapping, Type

import attr

from bento.constants import RESOURCE_PATH
from bento.formatter.base import FindingsMap, Formatter
from bento.formatter.json import Json
from bento.git import url
from bento.network import no_auth_post


class Serializer(ABC):
    @abstractmethod
    def serialize(
        self, config: Mapping[str, Any], findings: FindingsMap
    ) -> Mapping[str, Any]:
        pass


class Circle(Serializer):
    def serialize(
        self, config: Mapping[str, Any], findings: FindingsMap
    ) -> Mapping[str, Any]:
        return {
            "meta": {"repo_url": url()},
            "config": config,
            "findings": Json.to_py(findings),
        }


SCHEMATA: Mapping[str, Type[Serializer]] = {"circle": Circle}

DEFAULT_TIMEOUT_SECONDS = 2.0


class IterColl(collections.Collection):
    def __init__(self, length: int, *iters: Iterator[str]) -> None:
        self.length = length
        self.iters = iters

    def __contains__(self, __x: object) -> bool:
        return False

    def __iter__(self) -> Iterator[str]:
        return itertools.chain(*self.iters)

    def __len__(self) -> int:
        return self.length


@attr.s
class Reporter(Formatter):
    def inner(self) -> Serializer:
        key: str = self.config.get("schema", "circle")
        if key not in SCHEMATA:
            raise Exception(f"Unknown reporter schema '{key}'")
        return SCHEMATA[key]()

    def url(self) -> str:
        if "url" not in self.config:
            raise Exception(
                f"Config for reporter formatter is missing 'url'. Please add 'url: URL' to the configuration for 'reporter' to {RESOURCE_PATH}"
            )
        return self.config["url"]

    def submit(self, url: str, data: Mapping) -> str:
        try:
            r = no_auth_post(url, json=data, timeout=self.timeout())
            r.raise_for_status()
            return "Successfully submitted report"
        except Exception:
            logging.exception(f"error posting to {url}")
            return "Failed to post report"

    def timeout(self) -> float:
        return self.config.get("timeout", DEFAULT_TIMEOUT_SECONDS)

    def dump(self, findings: FindingsMap) -> Collection[str]:
        blob = self.inner().serialize(self.context.config, findings)
        url = self.url()
        return IterColl(
            2,
            (s for s in [f"Publishing results to {url}..."]),
            (self.submit(url, d) for d in [blob]),
        )



import json
from typing import Iterable, List, Type

from semantic_version import SimpleSpec

from bento.parser import Parser
from bento.tool import output, runner
from bento.violation import Violation

"""
Input example:

{
    "/path/to/file.py": [
        {
            "code": "DUO105",
            "filename": "/path/to/file.py",
            "line_number": 24,
            "column_number": 1,
            "text": "use of \"exec\" is insecure",
            "physical_line": "exec('foo')\n"
        }
    ]
}
"""

DLINT_TO_BENTO = {"DUO138": "regular-expression-catastrophic-backtracking"}


class DlintParser(Parser[str]):
    SEVERITY = {"LOW": 0, "MEDIUM": 1, "HIGH": 2}

    @staticmethod
    def _get_link(code: str) -> str:
        return "https://github.com/dlint-py/dlint/blob/master/docs/linters/{}.md".format(
            code
        )

    def parse(self, tool_output: str) -> List[Violation]:
        results = json.loads(tool_output)

        violations = [
            Violation(
                check_id=DLINT_TO_BENTO[result["code"]],
                tool_id=DlintTool.TOOL_ID,
                path=self.trim_base(filename),
                severity=DlintParser.SEVERITY["MEDIUM"],
                line=result["line_number"],
                column=result["column_number"],
                message=result["text"],
                syntactic_context=(result["physical_line"] or "").rstrip(),
                link=self._get_link(result["code"]),
            )
            for filename, file_results in results.items()
            for result in file_results
        ]

        return violations


class DlintTool(runner.Python, output.Str):
    TOOL_ID = "dlint"
    VENV_DIR = "dlint"
    PROJECT_NAME = "Python"
    PACKAGES = {"dlint": SimpleSpec("~=0.10.2"), "flake8-json": SimpleSpec("~=19.8.0")}

    @property
    def parser_type(self) -> Type[Parser]:
        return DlintParser

    @classmethod
    def tool_id(self) -> str:
        return DlintTool.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "A tool for encouraging best coding practices and helping ensure Python code is secure"

    @classmethod
    def venv_subdir_name(cls) -> str:
        return DlintTool.VENV_DIR

    @property
    def project_name(self) -> str:
        return DlintTool.PROJECT_NAME

    def select_clause(self) -> str:
        return f"--select={','.join(DLINT_TO_BENTO.keys())}"

    def run(self, paths: Iterable[str]) -> str:
        cmd = [
            "python",
            str(self.venv_dir() / "bin" / "flake8"),
            self.select_clause(),
            "--format=json",
            "--isolated",
            *paths,
        ]
        return self.venv_exec(cmd, check_output=False)



import json
import logging
import os
from pathlib import Path
from typing import Iterable, Optional

import attr
import pymmh3 as mmh3

from bento import __version__ as BENTO_VERSION


@attr.s
class RunCache:
    """
        Acts as a local cache for tool run output

        Different tools can be accessed concurrently, but cache access
        is not threadsafe if multiple threads access the same tool.
    """

    cache_dir: Path = attr.ib(converter=Path)
    force_invalidate_cache: bool = os.environ.get("BENTO_USE_CACHE", "1") == "0"

    def __cache_metadata_path(self, tool_id: str) -> Path:
        """
            Returns name of file that cache results metadata for a given tool
        """
        return self.cache_dir / f"{tool_id}-meta.json"

    def __cache_data_path(self, tool_id: str) -> Path:
        """
            Returns name of file that cache results would be contained in
        """
        return self.cache_dir / f"{tool_id}.data"

    def _modified_hash(self, paths: Iterable[Path]) -> str:
        """
        Returns a hash of the recursive mtime of a path.

        Any modification of a file within this tree (that does not match an ignore pattern)
        will change the hash.
        """

        # No matter settings of .bentoignore, these are always excluded
        files_and_times = ((str(p), p.stat().st_mtime_ns) for p in paths if p.exists())
        h = 0
        for f, m in files_and_times:
            h ^= mmh3.hash128(f"{f}:{m}")

        return format(h, "x")

    def __cleanup(self, tool_id: str) -> None:
        """
            Delete all state relevant for cacheing tool_id
        """
        cache_metadata_path = self.__cache_metadata_path(tool_id)
        cache_data_path = self.__cache_data_path(tool_id)

        # Silently delete file if exists
        # note that checking for file before deletion
        # has a TOCTOU race so will need a try-catch anyway
        try:
            cache_metadata_path.unlink()
        except OSError:
            pass

        try:
            cache_data_path.unlink()
        except OSError:
            pass

    def wipe(self) -> None:
        try:
            self.cache_dir.unlink()
        except OSError:
            pass

    def get(self, tool_id: str, paths: Iterable[Path]) -> Optional[str]:
        """
            Returns stored run output if it exists in local run cache and the
            cache entry is still valid (files have not been modified since caching)

            Returns None if no such cache entry is found
        """
        cache_metadata_path = self.__cache_metadata_path(tool_id)
        cache_data_path = self.__cache_data_path(tool_id)

        if not (cache_metadata_path.exists() and cache_data_path.exists()):
            self.__cleanup(tool_id)
            return None

        with cache_metadata_path.open() as file:
            try:
                metadata = json.load(file)
            except json.JSONDecodeError as e:
                logging.error(f"Failed to parse tool {tool_id} cache metadata as json")
                logging.error(e.msg, e.doc, e.pos)
                self.__cleanup(tool_id)
                return None

        cache_hash = metadata.get("hash")
        cache_paths = {Path(p) for p in metadata.get("paths")}
        cache_bento_version = metadata.get("version")

        if (
            self.force_invalidate_cache
            or cache_paths != set(paths)
            or cache_bento_version != BENTO_VERSION
            or cache_hash != self._modified_hash(paths)
        ):
            logging.warning(f"Invalidating cache for {tool_id}")
            self.__cleanup(tool_id)
            return None

        return cache_data_path.read_text()

    def put(self, tool_id: str, paths: Iterable[Path], raw_results: str) -> None:
        """
            Caches raw_results as the output of running TOOL_ID on PATHS

            Note that RunCache.get assumed paths is not None so should be changed
            if PATHS here is nullable
        """
        self.__cleanup(tool_id)

        self.cache_dir.mkdir(parents=True, exist_ok=True)
        cache_metadata_path = self.__cache_metadata_path(tool_id)
        cache_data_path = self.__cache_data_path(tool_id)

        # Data should be written before metadata
        cache_data_path.write_text(raw_results)
        hsh = self._modified_hash(paths)

        metadata = {
            "paths": [str(p) for p in paths],
            "hash": hsh,
            "version": BENTO_VERSION,
        }

        with cache_metadata_path.open("w") as file:
            json.dump(metadata, file)



import logging
import os
import sys
import time
from distutils.util import strtobool
from pathlib import Path
from typing import Optional

import click
from packaging.version import InvalidVersion, Version

import bento.constants as constants
import bento.network
from bento.commands import archive, check, disable, enable, init, register
from bento.context import Context
from bento.error import InvalidRegistrationException, OutdatedPythonException


def _setup_logging() -> None:
    os.makedirs(os.path.dirname(constants.DEFAULT_LOG_PATH), exist_ok=True)
    logging.basicConfig(
        filename=constants.DEFAULT_LOG_PATH,
        level=logging.DEBUG,
        filemode="w",
        format="[%(levelname)s] %(relativeCreated)s %(name)s:%(module)s - %(message)s",
    )
    logging.info(
        f"Environment: stdout.isatty={sys.stdout.isatty()} stderr.isatty={sys.stderr.isatty()} stdin.isatty={sys.stdin.isatty()}"
    )
    # Very noisy logging from urllib3
    logging.getLogger("urllib3").setLevel(logging.WARNING)


def _is_test() -> bool:
    test_var = os.getenv(constants.BENTO_TEST_VAR, "0")
    try:
        return strtobool(test_var) == 1
    except ValueError:
        return False


def _get_version_from_cache(version_cache_path: Path) -> Optional[Version]:
    """
        Reads version cache file and returns Version stored in it
        if time cache was written was within the last day.

        Returns None if version cache is invalid for any reason
        or file does not exist
    """
    now = time.time()
    if version_cache_path.is_file():
        with version_cache_path.open() as f:
            timestamp_str = f.readline().strip()
            latest_version_str = f.readline().strip()
            try:
                latest_version = Version(latest_version_str)
            except InvalidVersion:
                logging.debug(
                    f"Version Cache invalid version string: {latest_version_str}"
                )
                return None

            try:
                # Treat time as integer seconds so no need to deal with str float conversion
                timestamp = int(timestamp_str)
            except ValueError:
                logging.debug(f"Version Cache invalid timestamp: {timestamp_str}")
                return None

            if now - timestamp > 86400:
                logging.debug(f"Version Cache expired {timestamp}:{now}")
                return None

            logging.debug(f"Version Cache returning {latest_version}")
            return latest_version
    logging.debug("Version Cache does not exist")
    return None


def _get_latest_version(version_cache_path: Path) -> Optional[Version]:
    """
        Return latest Version of bento-cli available.

        Checks local version cache to save from making network call
        but if cache is invalid makes network call and writes to cache.

        Returns None if cache is invalid and network call fails for
        any reason
    """
    latest_version = _get_version_from_cache(version_cache_path)
    if latest_version is None:
        latest_version_str, _ = bento.network.fetch_latest_version()
        if latest_version_str is None:
            # Request timed out or invalid
            return None

        try:
            latest_version = Version(latest_version_str)
        except InvalidVersion:
            # latest version str from server incorrect
            return None

        # Write to version cache
        with version_cache_path.open("w") as f:
            # Integer time so no need to deal with str float conversions
            f.write(f"{int(time.time())}\n")
            f.write(latest_version_str)

    return latest_version


def _is_running_latest() -> bool:
    """
        Returns True if current version of bento is latest version available
    """
    latest_version = _get_latest_version(constants.GLOBAL_VERSION_CACHE_PATH)
    current_version = Version(_get_version())
    logging.info(
        f"Current bento version is {current_version}, latest is {latest_version}"
    )
    if latest_version and current_version < latest_version:
        return False
    return True


def _get_version() -> str:
    """Get the current r2c-cli version based on __init__"""
    from bento import __version__

    return __version__


def _is_running_supported_python3() -> bool:
    python_major_v = sys.version_info.major
    python_minor_v = sys.version_info.minor
    logging.info(f"Python version is ({python_major_v}.{python_minor_v})")
    return python_major_v >= 3 and python_minor_v >= 6


def make_cli(prog_name: str, headless: bool = False) -> click.Command:
    @click.group(
        epilog=f"To get help for a specific command, run `{prog_name} COMMAND --help`"
    )
    @click.help_option("-h", "--help")
    @click.version_option(
        prog_name=prog_name, version=_get_version(), message="%(prog)s/%(version)s"
    )
    @click.option(
        "--base-path",
        help=f"Path to the project to run {prog_name} in.",
        type=click.Path(exists=True, file_okay=False),
        default=None,
        hidden=True,
    )
    @click.option(
        "--cache-path",
        type=click.Path(writable=True, file_okay=False),
        help="Path to cache directory. Tool metadata and results will be cached here.",
        default=None,
        hidden=True,
    )
    @click.option(
        "--agree",
        is_flag=True,
        help="Automatically agree to terms of service.",
        default=False,
        hidden=headless,
    )
    @click.option(
        "--email",
        type=str,
        help="Email address to use while running this command without global configs e.g. in CI",
        default=None,
        hidden=headless,
    )
    @click.option("--headless", is_flag=True, default=False, hidden=True)
    @click.pass_context
    def cli(
        ctx: click.Context,
        base_path: Optional[str],
        cache_path: Optional[str],
        agree: bool,
        headless: bool,
        email: Optional[str],
    ) -> None:
        _setup_logging()
        is_init = ctx.invoked_subcommand == "init"
        ctx.help_option_names = ["-h", "--help"]
        extra = {"base_path": base_path} if base_path is not None else {}
        ctx.obj = Context(
            is_init=is_init,
            email=email,
            headless=headless,
            prog_name=prog_name,
            cache_path=cache_path,
            **extra,  # type: ignore
        )
        if not _is_running_supported_python3():
            raise OutdatedPythonException()

        if not headless:
            registrar = register.Registrar(ctx, agree, email=email)
            if not registrar.verify():
                raise InvalidRegistrationException()

        if not _is_running_latest() and not _is_test():
            logging.warning(f"{prog_name} client is outdated")
            click.secho(constants.UPGRADE_WARNING_OUTPUT, err=True)

    cli.add_command(check.make_check(prog_name, headless))
    cli.add_command(archive.make_archive(prog_name, headless))

    if not headless:
        cli.add_command(init.init)
        cli.add_command(enable.enable)
        cli.add_command(disable.disable)

    return cli



from typing import Set

import click

from bento.commands.autocomplete import install_autocomplete
from bento.commands.autorun import install_autorun
from bento.commands.ci import install_ci
from bento.config import (
    ToolCommand,
    get_disabled_checks,
    get_valid_tools,
    update_ignores,
    update_tool_run,
)
from bento.context import Context
from bento.decorators import with_metrics
from bento.util import echo_success


@click.group()
def enable() -> None:
    """
    Turn ON a Bento feature for this project.

    For example, to enable autorun:

        $ bento enable autorun

    To enable a non-default tool (like `eslint`):

        $ bento enable tool eslint
    """


@enable.command(
    cls=ToolCommand,
    short_help="Specify a tool to enable.",
    help_summary="Turn ON a tool.",
)
@click.argument("tool", type=str, nargs=1, autocompletion=get_valid_tools)
@with_metrics
@click.pass_obj
def tool(context: Context, tool: str) -> None:
    """
    Turn ON a tool.

    If the tool was previously enabled, the tool's previous
    settings will be used. If no configuration exists, smart defaults will
    be applied.

    See `bento enable --help` for more details.
    """
    update_tool_run(context, tool, True)
    echo_success(f"{tool} enabled")


@enable.command(short_help="Specify a check to enable.")
@click.argument("tool", type=str, nargs=1, autocompletion=get_valid_tools)
@click.argument("check", type=str, nargs=1, autocompletion=get_disabled_checks)
@with_metrics
@click.pass_obj
def check(context: Context, tool: str, check: str) -> None:
    """
    Turn ON a check.

    Visit bento.dev/checks to learn about Bento's specialty checks.
    """

    def remove(ignores: Set[str]) -> None:
        if check in ignores:
            ignores.remove(check)

    update_ignores(context, tool, remove)
    echo_success(f"'{check}' enabled for '{tool}'")


enable.add_command(install_autorun)
enable.add_command(install_autocomplete)
enable.add_command(install_ci)



from typing import Type

from bento.extra.base_sgrep import BaseSgrepParser, BaseSgrepTool
from bento.parser import Parser


class R2cCheckRegistryParser(BaseSgrepParser):
    @classmethod
    def tool_id(cls) -> str:
        return R2cCheckRegistryTool.tool_id()


class R2cCheckRegistryTool(BaseSgrepTool):
    TOOL_ID = "r2c.registry.latest"
    CONFIG = "https://r2c.dev/default-r2c-checks"
    PARSER = R2cCheckRegistryParser

    @classmethod
    def tool_id(cls) -> str:
        return cls.TOOL_ID

    @property
    def config_str(self) -> str:
        return self.CONFIG

    @property
    def parser_type(self) -> Type[Parser]:
        return self.PARSER



from __future__ import unicode_literals

import itertools
import logging
import os
import os.path
import re
import shutil
import signal
import subprocess
import sys
import threading
import types
from functools import partial
from importlib import import_module
from pathlib import Path
from textwrap import wrap as py_wrap
from typing import (
    IO,
    Any,
    Callable,
    Collection,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Pattern,
    TextIO,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import psutil
import yaml
from click.termui import secho, style
from frozendict import frozendict

import bento.constants as constants

EMPTY_DICT = frozendict({})
MAX_PRINT_WIDTH = 80
MIN_PRINT_WIDTH = 45
ANSI_WIDTH = 4  # number of characters to emit an ANSI control code
LEADER_CHAR = "․"
SETUP_TEXT = "🍜 Setting up"
SETUP_WIDTH = len(SETUP_TEXT)
PROGRESS_TEXT = "🍤 Running".ljust(SETUP_WIDTH, " ")
DONE_TEXT = "🍱 Done".ljust(SETUP_WIDTH, " ")
SKIP_TEXT = "👋 Skipped".ljust(SETUP_WIDTH, " ")
RESET_TEXT = "".ljust(SETUP_WIDTH + 1, "\b")  # +1 for emoji width

OSC_8 = "\x1b]8;;"
BEL = "\x07"

LINK_PRINTER_PATTERN = re.compile("(iterm2|gnome-terminal)", re.IGNORECASE)
LINK_WIDTH = 2 * len(OSC_8) + 2 * len(BEL)

AutocompleteSuggestions = List[Union[str, Tuple[str, str]]]


class Colors:
    LINK = "bright_blue"
    ERROR = "red"
    WARNING = "yellow"
    SUCCESS = "green"


def _calculate_print_width() -> int:
    term_width, _ = shutil.get_terminal_size((MAX_PRINT_WIDTH, 0))
    return max(min(MAX_PRINT_WIDTH, term_width), MIN_PRINT_WIDTH)


PRINT_WIDTH = _calculate_print_width()


def read_global_config() -> Optional[Dict[str, Any]]:
    if not os.path.exists(constants.GLOBAL_CONFIG_PATH):
        return None

    with open(constants.GLOBAL_CONFIG_PATH, "r") as yaml_file:
        try:
            return yaml.safe_load(yaml_file)
        except Exception:
            logging.warning("Invalid global config file found")
            return None


def persist_global_config(global_config: Dict[str, Any]) -> None:
    os.makedirs(constants.GLOBAL_RESOURCE_PATH, exist_ok=True)
    with open(constants.GLOBAL_CONFIG_PATH, "w+") as yaml_file:
        yaml.safe_dump(global_config, yaml_file)

    logging.info(f"Updated user configs at {constants.GLOBAL_CONFIG_PATH}.")


def fetch_line_in_file(path: Path, line_number: int) -> Optional[str]:
    """
    `line_number` is one-indexed! Returns the line if it can be found, returns None if the path doesn't exist
    """
    if not path.exists():
        return None
    with path.open(buffering=1) as fin:  # buffering=1 turns on line-level reads
        return next(itertools.islice(fin, line_number - 1, line_number), None)


def for_name(name: str) -> Type:
    """
    Reflectively obtains a type from a python identifier

    E.g.
        for_name("bento.extra.eslint.EslintTool")
    returns the EslintTool type

    Parameters:
        name (str): The type name, as a python fully qualified identifier
    """
    module_name, class_name = name.rsplit(".", 1)
    mod = import_module(module_name)
    return getattr(mod, class_name)


def is_child_process_of(pattern: Pattern) -> bool:
    """
    Returns true iff this process is a child process of a process whose name matches pattern
    """
    me = psutil.Process()
    parents = me.parents()
    matches = iter(0 for p in parents if pattern.search(p.name()))
    return next(matches, None) is not None


M = TypeVar("M", covariant=True)


class Memo(Generic[M]):
    """
    Creates a (thread-safe) lazily loaded value
    """

    def __init__(self, load: Callable[[], M]) -> None:
        self.load = load
        self._lock = threading.Lock()
        self._value: Optional[M] = None

    @property
    def value(self) -> M:
        with self._lock:
            if self._value is None:
                self._value = self.load()
            return self._value


DO_PRINT_LINKS = is_child_process_of(LINK_PRINTER_PATTERN)


_T = TypeVar("_T")


def batched(it: Iterable[_T], max_len: int) -> Iterator[Iterator[_T]]:
    """
    Batches an iterator in iterators of a maximum length

    :param it: The iterator to batch
    :param max_len: The maximum length
    """
    return (
        (y for _, y in x)
        for _, x in itertools.groupby(enumerate(it), key=lambda x: int(x[0] / max_len))
    )


def less(
    output: Collection[Collection[str]], pager: bool = True, overrun_pages: int = 0
) -> None:
    """
    Possibly prints a string through less.

    Parameters:
        pager: If false, the string is always echoed directly to stdout
        overrun_pages: Minimum number of pages in output before paging is triggered (paging is never triggered if
                       less than or equal to 0)
    """
    use_echo = False
    text = (line for o in output for line in o)
    text_len = sum(len(o) for o in output)

    # In order to prevent an early pager exit from killing the CLI,
    # we must both ignore the resulting SIGPIPE and BrokenPipeError
    def drop_sig(signal: int, frame: Optional[types.FrameType]) -> None:
        pass

    if not pager or not sys.stdout.isatty():
        use_echo = True

    if not use_echo:
        _, height = shutil.get_terminal_size()
        if text_len < height * overrun_pages:
            use_echo = True

    if use_echo:
        for t in text:
            secho(t)
    else:
        # NOTE: Using signal.SIG_IGN here DOES NOT IGNORE the resulting SIGPIPE
        signal.signal(signal.SIGPIPE, drop_sig)
        try:
            process = subprocess.Popen(["less", "-r"], stdin=subprocess.PIPE)
            stdin = cast(IO[bytes], process.stdin)
            for ix, t in enumerate(text):
                stdin.write(bytearray(t, "utf8"))
                if ix != text_len - 1:
                    stdin.write(bytearray("\n", "utf8"))
            process.communicate()
        except BrokenPipeError:
            pass
        finally:
            signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def wrap(text: str, extra: int = 0) -> str:
    """
    Wraps text to (one character less than) the screen print width

    :param text: The text to wrap
    :param extra: Any extra width to apply
    """
    return "\n".join(py_wrap(text, PRINT_WIDTH - 1 + extra))


def echo_styled(
    text: str, icon: str, log_fn: Callable, fg: str, indent: str = ""
) -> None:
    """
    Print operational messages to stderr with styling.
    Comes with preconfigured functions for common styles:
    `echo_error`, `echo_warning`, and `echo_success`


    :param text: Text to print, will be wrapped line-by-line.
    :param icon: An icon, emoji or otherwise, to prefix the first line with.
    :param log_fn: A function to call with the raw text.
                   Meant to write to log files in the appropriate format.
    :param fg: A click-compatible color name to use as text color.
    :param indent: A string to prepend to every line, such as "  " or "... "

    >>> echo_error("ERROR\nTask failed successfully.", indent="  ")
      ✘ ERROR
      Task failed successfully.
    """
    log_fn(text)
    prefix = f"{icon} "
    lines = text.splitlines()
    for line in lines:
        secho(wrap(f"{indent}{prefix}{line}"), fg=fg, err=True)
        prefix = ""


echo_error = partial(echo_styled, icon="✘", log_fn=logging.error, fg=Colors.ERROR)
echo_warning = partial(echo_styled, icon="⚠", log_fn=logging.warning, fg=Colors.WARNING)
echo_success = partial(echo_styled, icon="✔", log_fn=logging.info, fg=Colors.SUCCESS)


def echo_box(text: str) -> None:
    """
    Prints text bold, in a header box

    By default, the box is PRINT_WIDTH characters wide, unless the text is too
    long for the box, in which case the box is extended to fit.
    """
    lines = text.split("\n")
    max_len = max(len(l) for l in lines)
    max_len = max(PRINT_WIDTH - 4, max_len)
    hrule = "".ljust(max_len + 2, "─")
    echo_newline()
    secho(f"╭{hrule}╮", err=True)
    for l in lines:
        p = style(f"{l:^{max_len}s}", bold=True)
        secho(f"│ {p} │", err=True)
    secho(f"╰{hrule}╯", err=True)


def echo_newline() -> None:
    """
    Prints an informational newline (printed to stderr)
    """
    secho("", err=True)


def echo_styles(*parts: str) -> None:
    """
    Echoes concatenated styled parts to stderr

    :param parts: List of strings to print
    """
    for p in parts:
        if isinstance(p, str):
            secho(p, nl=False, err=True)
    echo_newline()


def echo_next_step(desc: str, cmd: str) -> None:
    """
    Echoes a 'next step' to perform, with styling.

    E.g.

      ◦ To archive results, run $ bento archive

    :param desc: The step description
    :param cmd: The command that the user should run
    """
    echo_styles("◦ ", style(f"{desc}, run $ ", dim=True), cmd)


def echo_progress(text: str, extra: int = 0, skip: bool = False) -> Callable[[], None]:
    """
    Prints a binary in-progress / done bar

    Usage example:
      mark_done = echo_progress("Installing foo")
      install_foo()
      mark_done()

    :param extra: Number of unprinted characters in text (each ANSI code point is 4 characters)
    :param skip: If true, "Skipped" is printed instead, and callback is a no-op
    """
    width = PRINT_WIDTH - 3 - SETUP_WIDTH + ANSI_WIDTH + extra
    logging.info(text)
    leader = style("".ljust(width - len(text), LEADER_CHAR), dim=True)

    if skip:
        secho(f"{text}{leader} {style(SKIP_TEXT, dim=True)}", err=True, dim=True)
        return lambda: None
    else:
        secho(f"{text}{leader} {SETUP_TEXT}", nl=False, err=True, dim=True)
        return lambda: secho(f"{RESET_TEXT}{DONE_TEXT}", err=True, dim=True)


def render_link(
    text: str,
    href: Optional[str],
    print_alternative: bool = True,
    width: Optional[int] = None,
    pipe: TextIO = sys.stdout,
) -> str:
    """
    Prints a clickable hyperlink output if in a tty; otherwise just prints a text link

    :param text: The link anchor text
    :param href: The href, if exists
    :param print_alternative: If true, only emits link if OSC8 links are supported, otherwise prints href after text
    :param width: Minimum link width
    :param pipe: The text IO via which this link will be emitted
    :return: The rendered link
    """
    is_rendered = False
    if href:  # Don't render if href is None or empty
        if pipe.isatty() and DO_PRINT_LINKS:
            text = f"{OSC_8}{href}{BEL}{text}{OSC_8}{BEL}"
            is_rendered = True
            if width:
                width += LINK_WIDTH + len(href)
        elif print_alternative:
            text = f"{text} {href}"

    if width:
        text = text.ljust(width)

    # Coloring has to occur after justification
    if is_rendered:
        text = style(text, fg=Colors.LINK)

    return text


def file_has_text(file: Path, text: str) -> bool:
    """
    Returns if a file contains a pattern
    """
    with file.open() as fd:
        return any(text in l for l in fd)


def append_text_to_file(file: Path, text: str) -> None:
    """
    Adds text to a file
    """
    with file.open("a") as fd:
        fd.write(f"\n{text}\n")



from typing import Iterable, Type

from bento.extra.bandit import BanditTool
from bento.extra.boto3 import Boto3Tool
from bento.extra.click import ClickTool
from bento.extra.dlint import DlintTool
from bento.extra.eslint import EslintTool
from bento.extra.flake8 import Flake8Tool
from bento.extra.flask import FlaskTool
from bento.extra.gosec import GosecTool
from bento.extra.grep import GrepTool
from bento.extra.hadolint import HadolintTool
from bento.extra.jinjalint import JinjalintTool
from bento.extra.pyre import PyreTool
from bento.extra.r2c_check_registry import R2cCheckRegistryTool
from bento.extra.requests import RequestsTool
from bento.extra.sgrep import SgrepTool
from bento.extra.shellcheck import ShellcheckTool
from bento.tool import Tool

TOOLS: Iterable[Type[Tool]] = [
    BanditTool,
    Boto3Tool,
    ClickTool,
    DlintTool,
    EslintTool,
    FlaskTool,
    Flake8Tool,
    GosecTool,
    GrepTool,
    HadolintTool,
    JinjalintTool,
    # PyreTool,  # Pyre does not work with explicit paths
    RequestsTool,
    SgrepTool,
    ShellcheckTool,
    R2cCheckRegistryTool,
]

EXPERIMENTAL_TOOLS = {PyreTool, ClickTool, SgrepTool, GrepTool}



from typing import Type

from semantic_version import SimpleSpec

from bento.extra.flake8 import Flake8Parser, Flake8Tool
from bento.parser import Parser
from bento.tool import output

PREFIX = "r2c-click-"


class ClickParser(Flake8Parser):
    @staticmethod
    def id_to_link(check_id: str) -> str:
        page = ClickParser.id_to_name(check_id)
        return f"https://bento.dev/checks/click/{page}/"

    @classmethod
    def id_to_name(cls, check_id: str) -> str:
        return check_id.replace(PREFIX, "")

    @staticmethod
    def tool() -> Type[output.Str]:
        return ClickTool


class ClickTool(Flake8Tool):
    TOOL_ID = "r2c.click"  # to-do: versioning?
    VENV_DIR = "click"
    PACKAGES = {
        "flake8": SimpleSpec("~=3.7.0"),
        "flake8-json": SimpleSpec("~=19.8.0"),
        "flake8-click": SimpleSpec("==0.3.1"),
    }

    @property
    def parser_type(self) -> Type[Parser]:
        return ClickParser

    @classmethod
    def tool_id(cls) -> str:
        return cls.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Checks for the Python Click framework"

    def select_clause(self) -> str:
        return "--select=r2c-click"



import json
from typing import Any, Collection, Mapping, Sequence

from bento.formatter.base import FindingsMap, Formatter


class Json(Formatter):
    """Formats output as a single JSON blob."""

    @staticmethod
    def to_py(findings: FindingsMap) -> Sequence[Mapping[str, Any]]:
        return [
            {
                "tool_id": violation.tool_id,
                "check_id": violation.check_id,
                "line": violation.line,
                "column": violation.column,
                "message": violation.message,
                "severity": violation.severity,
                "path": violation.path,
            }
            for violations in findings.values()
            for violation in violations
        ]

    def dump(self, findings: FindingsMap) -> Collection[str]:
        return [json.dumps(self.to_py(findings))]



import logging
import resource
import subprocess
from abc import ABC, abstractmethod
from pathlib import Path
from time import time
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Pattern,
    Set,
    Type,
    TypeVar,
)

import attr

from bento.base_context import BaseContext
from bento.parser import Parser
from bento.result import from_cache_repr, to_cache_repr
from bento.util import batched
from bento.violation import Violation

R = TypeVar("R")
"""Generic return type"""

JsonR = List[Dict[str, Any]]
"""Return type for tools with a JSON representation"""


MIN_RESERVED_ARGS = 128
"""Number of argument positions reserved for non-file command arguments (e.g. rule ignores)"""


# Note: for now, every tool *HAS* to directly inherit from this, even if it
# also inherits from JsTool or PythonTool. This is so we can list all tools by
# looking at subclasses of Tool.
@attr.s
class Tool(ABC, Generic[R]):
    """The base class for all tool plugins"""

    context = attr.ib(type=BaseContext)

    @property
    def base_path(self) -> Path:
        """Returns the base path from which this tool should run"""
        return self.context.base_path

    @property
    def config(self) -> Dict[str, Any]:
        """Returns this tool's configuration"""
        return self.context.config["tools"][self.tool_id()] or {}

    def parser(self) -> Parser[R]:
        """Returns this tool's parser"""
        return self.parser_type(self.base_path)

    @property
    @abstractmethod
    def parser_type(self) -> Type[Parser[R]]:
        """Returns this tool's parser type"""
        pass

    @classmethod
    @abstractmethod
    def tool_id(cls) -> str:
        """Returns this tool's string ID"""
        pass

    @classmethod
    @abstractmethod
    def tool_desc(cls) -> str:
        """Returns a description of what this tool tests"""
        pass

    @property
    @abstractmethod
    def project_name(self) -> str:
        """Returns this tool's human friendly project name"""
        pass

    @property
    @abstractmethod
    def file_name_filter(self) -> Pattern:
        """Returns a pattern that determines whether a terminal path should be run through this tool"""
        pass

    @property
    def shebang_pattern(self) -> Optional[Pattern]:
        return None

    @abstractmethod
    def setup(self) -> None:
        """
        Runs all code necessary to install this tool

        Code runs inside virtual environment.

        Parameters:
            config (dict): The tool configuration

        Raises:
            CalledProcessError: If setup fails
        """
        pass

    @abstractmethod
    def run(self, files: Iterable[str]) -> R:
        """
        Runs this tool, returning its results

        Code runs inside virtual environment.

        Parameters:
            config (dict): The tool configuration

        Raises:
            CalledProcessError: If execution fails
        """
        pass

    @abstractmethod
    def matches_project(self, files: Iterable[Path]) -> bool:
        """
        Returns true if and only if this project should use this tool
        """
        pass

    def can_use_cache(self) -> bool:
        """
        Returns true if this tool can use cached results

        This should return false if any datum changes that affects caching, other than the file set or the
        Bento version.
        """
        return True

    def extra_cache_paths(self) -> List[Path]:
        """
        Returns extra paths beyond the checked paths whose change should invalidate the cache

        For example, if the tool has a filesystem configuration file, it should be returned here.
        """
        return []

    def project_has_file_paths(self, files: Iterable[Path]) -> bool:
        """
        Returns true iff any unignored files matches at least one extension
        """
        return any(self.filter_paths(files))

    def _file_contains_shebang_pattern(self, file_path: Path) -> bool:
        """
            Checks if the first line of file_path matches self.shebang_pattern.

            Returns False if a "first line" does not make sense for file_path
            i.e. file_path is a binary or is an empty file or doesnt exist
        """
        assert self.shebang_pattern is not None

        if not file_path.is_file():
            # File doesnt exist or is a dir
            return False

        with open(file_path) as file:
            try:
                line: str = next(file)
            except StopIteration:  # Empty file
                return False
            except UnicodeDecodeError:  # Binary file
                return False

        return self.shebang_pattern.match(line) is not None

    def filter_paths(self, paths: Iterable[Path]) -> Set[Path]:
        """
        Filters a list of paths to those that should be analyzed by this tool

        In it's default behavior, this method:
          - Filters terminal paths (files) that match file_name_filter.
          - Filters non-terminal paths (directories) that include at least one matching path.
          - If shebang_pattern is defined filters files that do not match file_name_filter but
            said file's first line matches shebang_pattern

        Parameters:
            paths (list): List of candidate paths

        Returns:
            A set of valid paths
        """
        abspaths = [p.resolve() for p in paths]
        to_run = {
            p
            for p in abspaths
            if self.file_name_filter.match(p.name)
            or (self.shebang_pattern and self._file_contains_shebang_pattern(p))
        }
        return to_run

    def execute(self, command: List[str], **kwargs: Any) -> subprocess.CompletedProcess:
        """
        Delegates to subprocess.run() on this tool's base path

        Raises:
            CalledProcessError: If execution fails and check is True
        """
        new_args: Dict[str, Any] = {"cwd": self.base_path, "encoding": "utf8"}
        new_args.update(kwargs)
        cmd_args = (f"'{a}'" for a in command)
        logging.debug(f"{self.tool_id()}: Running: {' '.join(cmd_args)}")
        before = time()
        res = subprocess.run(command, **new_args)
        after = time()
        logging.debug(f"{self.tool_id()}: Command completed in {after - before:2f} s")
        return res

    @classmethod
    def max_batch_size(cls) -> int:
        """Returns the maximum number of files to run in a single batch"""
        # On UNIX, max argc is stack size / 4
        return int(resource.RLIMIT_STACK / 4) - MIN_RESERVED_ARGS

    def _get_findings_from_run(self, paths: Iterable[Path]) -> List[Violation]:
        """
        Returns findings by calling tool "run" method

        :param paths: Paths to run on
        :return:
        """
        paths_to_run = self.filter_paths(paths)
        if not paths_to_run:
            return []

        violations: List[Violation] = []

        for batch in batched(paths_to_run, self.max_batch_size()):
            path_list = [str(p) for p in batch]
            raw = self.run(path_list)
            try:
                violations += self.parser().parse(raw)
            except Exception as e:
                raise Exception(
                    f"Could not parse output of '{self.tool_id()}':\n{raw}", e
                )

        return violations

    def results(self, paths: List[Path], use_cache: bool = True) -> List[Violation]:
        """
        Runs this tool, returning all identified violations

        Code runs inside virtual environment.

        Before running tool, checks local RunCache for cached tool output and skips
        if said output is still usable

        Parameters:
            paths (list or None): If defined, an explicit list of paths to run on
            use_cache (bool): If True, checks for cached results

        Raises:
            CalledProcessError: If execution fails
        """
        if not paths:
            return []

        use_cache = use_cache and self.can_use_cache()

        logging.debug(f"Checking for local cache for {self.tool_id()}")
        cache_repr = self.context.cache.get(
            self.tool_id(), paths + self.extra_cache_paths()
        )
        if not use_cache or cache_repr is None:
            logging.debug(f"Cache entry invalid for {self.tool_id()}. Running Tool.")
            violations = self._get_findings_from_run(paths)
            if use_cache:
                self.context.cache.put(
                    self.tool_id(),
                    paths + self.extra_cache_paths(),
                    to_cache_repr(violations),
                )
        else:
            violations = from_cache_repr(cache_repr)

        ignore_set = set(self.config.get("ignore", []))
        filtered = [v for v in violations if v.check_id not in ignore_set]
        return filtered



import json
import re
from typing import Any, Dict, Iterable, List, Optional, Pattern, Type

import attr
from semantic_version import SimpleSpec

from bento.parser import Parser
from bento.tool import JsonR, output, runner
from bento.util import fetch_line_in_file
from bento.violation import Violation


class HadolintParser(Parser[JsonR]):
    def to_violation(self, result: Dict[str, Any]) -> Violation:
        start_line = result["line"]
        column = result["column"]
        check_id = result["code"]
        message = result["message"]
        path = result["file"]

        path = self.trim_base(path)

        level = result["level"]
        severity = 0
        if level == "error":
            severity = 2
        elif level == "warning":
            severity = 1

        if "DL" in check_id or check_id in ["SC2046", "SC2086"]:
            link = f"https://github.com/hadolint/hadolint/wiki/{check_id}"
        elif "SC" in check_id:
            link = f"https://github.com/koalaman/shellcheck/wiki/{check_id}"
        else:
            link = ""

        line_of_code = (
            fetch_line_in_file(self.base_path / path, start_line) or "<no source found>"
        )

        if check_id == "DL1000":
            message = "Dockerfile parse error. Invalid docker instruction."

        return Violation(
            tool_id=HadolintTool.TOOL_ID,
            check_id=check_id,
            path=path,
            line=start_line,
            column=column,
            message=message,
            severity=severity,
            syntactic_context=line_of_code,
            link=link,
        )

    def parse(self, results: JsonR) -> List[Violation]:
        return [self.to_violation(r) for r in results]


@attr.s
class HadolintTool(runner.Docker, output.Json):
    TOOL_ID = "hadolint"
    DOCKER_IMAGE = "hadolint/hadolint:v1.17.2-8-g65736cb"
    DOCKERFILE_FILTER = re.compile(".*Dockerfile.*", re.IGNORECASE)
    BINARY_SPEC = SimpleSpec("~=1.17.2")
    VERSION_MATCH = re.compile(r"Haskell Dockerfile Linter v([\d.]+).*")

    @property
    def binary_executable(self) -> Optional[str]:
        return "hadolint"

    @property
    def binary_version_pattern(self) -> Pattern:
        return self.VERSION_MATCH

    @property
    def binary_command(self) -> List[str]:
        # Docker command already includes executable
        return []

    @property
    def allowed_binary_versions(self) -> SimpleSpec:
        return self.BINARY_SPEC

    @property
    def parser_type(self) -> Type[Parser]:
        return HadolintParser

    @classmethod
    def tool_id(cls) -> str:
        return cls.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Finds bugs in Docker files (requires Docker)"

    @property
    def file_name_filter(self) -> Pattern:
        return self.DOCKERFILE_FILTER

    @property
    def project_name(self) -> str:
        return "Docker"

    @property
    def docker_image(self) -> str:
        return self.DOCKER_IMAGE

    @property
    def docker_command(self) -> List[str]:
        return ["hadolint", "--format", "json"]

    @property
    def remote_code_path(self) -> str:
        return "/mnt"

    def is_allowed_returncode(self, returncode: int) -> bool:
        return returncode == 0 or returncode == 1

    def run(self, files: Iterable[str]) -> JsonR:
        result = self.run_command(files).stdout
        return json.loads(result)



import logging
import time
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Tuple, Type

import attr

import bento.extra
import bento.formatter
from bento.base_context import BaseContext
from bento.error import EnabledToolNotFoundException, MultipleErrorsException
from bento.formatter import Formatter
from bento.tool import Tool

if TYPE_CHECKING:
    from bento.error import BentoException


@attr.s(repr=False)
class Context(BaseContext):
    _formatters = attr.ib(type=List[Formatter], default=None, init=False)
    _start = attr.ib(type=float, default=time.time(), init=False)
    _user_start = attr.ib(type=float, default=None, init=False)
    _user_duration = attr.ib(type=float, default=0.0, init=False)
    _timestamp = attr.ib(
        type=str, default=str(datetime.utcnow().isoformat("T")), init=False
    )
    _tool_inventory = attr.ib(type=Dict[str, Type[Tool]], init=False, default=None)
    _tools = attr.ib(type=Dict[str, Tool], init=False, default=None)
    _configured_tools = attr.ib(
        type=Optional[Dict[str, Tool]], init=False, default=None
    )
    _errors_on_exit = attr.ib(type=List["BentoException"], init=False, factory=list)

    def __repr__(self) -> str:
        return f"Context({self.base_path})"

    @property
    def formatters(self) -> List[Formatter]:
        if self._formatters is None:
            self._formatters = self._load_formatters()
        return self._formatters

    @property
    def tools(self) -> Dict[str, Tool]:
        """
        Returns all enabled tools
        """
        if self._tools is None:
            self._tools = self._load_enabled_tools()
        return self._tools

    @property
    def configured_tools(self) -> Dict[str, Tool]:
        """
        Returns all configured tools (whether or not they are disabled)
        """
        if self._configured_tools is None:
            self._configured_tools = self._load_configured_tools()
        return self._configured_tools

    @property
    def tool_inventory(self) -> Dict[str, Type[Tool]]:
        if self._tool_inventory is None:
            self._tool_inventory = self._load_tool_inventory()
        return self._tool_inventory

    @property
    def timestamp(self) -> str:
        return self._timestamp

    def elapsed(self) -> float:
        return time.time() - self._start

    def user_duration(self) -> Optional[float]:
        """
        Returns elapsed time in seconds since formatter was opened
        """
        return self._user_duration

    def tool(self, tool_id: str) -> Tool:
        """
        Returns a specific configured tool

        Raises:
            AttributeError: If the requested tool is not configured
        """
        tt = self.tools
        if tool_id not in tt:
            raise AttributeError(f"{tool_id} not one of {', '.join(tt.keys())}")
        return tt[tool_id]

    def start_user_timer(self) -> None:
        self._user_start = time.perf_counter()

    def stop_user_timer(self) -> None:
        if self._user_start is None:
            return
        self._user_duration += time.perf_counter() - self._user_start

    def error_on_exit(self, exc: "BentoException") -> None:
        """
        Queue up an exception to cause Bento to fail at the end of invocation.
        """
        self._errors_on_exit.append(exc)

    @property
    def on_exit_exception(self) -> Optional["BentoException"]:
        """
        The exception to raise at the end of Bento's invocation.

        :return: `None` if no errors were queued,
                 a `BentoException` if one error was queued,
                 or a `MultipleErrorsException` if several errors were queued.
        """
        if not self._errors_on_exit:
            return None
        elif len(self._errors_on_exit) == 1:
            return self._errors_on_exit[0]
        else:
            return MultipleErrorsException(self._errors_on_exit)

    def _load_tool_inventory(self) -> Dict[str, Type[Tool]]:
        """
        Loads all tools in the module into a dictionary indexed by tool_id
        """
        all_tools = {}
        for tt in bento.extra.TOOLS:
            tool_id = tt.tool_id()
            all_tools[tool_id] = tt
        logging.debug(f"Known tool IDs are {', '.join(all_tools.keys())}")
        return all_tools

    def _load_enabled_tools(self) -> Dict[str, Tool]:
        """
        Returns a list of this project's enabled tools
        These are the tools in the configuration file that do
        not have "run" option set to False
        """
        tools: Dict[str, Tool] = {}
        inventory = self.tool_inventory
        for tn, tool_config in self.config["tools"].items():
            if "run" in tool_config and not tool_config["run"]:
                continue

            ti = inventory.get(tn, None)
            if not ti:
                self.error_on_exit(EnabledToolNotFoundException(tool=tn))
                continue

            tools[tn] = ti(self)

        return tools

    def _load_configured_tools(self) -> Dict[str, Tool]:
        """
        Returns list of this project's configured tools (disabled and enabled)
        """
        tools: Dict[str, Tool] = {}
        inventory = self.tool_inventory
        for tn in self.config["tools"].keys():
            ti = inventory.get(tn, None)
            if not ti:
                # skip, we will only error if the tool is enabled
                continue

            tools[tn] = ti(self)

        return tools

    def _load_formatters(self) -> List[Formatter]:
        """
        Returns this project's configured formatter
        """
        if "formatter" not in self.config:
            return [bento.formatter.stylish.Stylish(self, {})]
        else:
            FormatterConfig = Dict[str, Any]
            FormatterSpec = Tuple[str, FormatterConfig]

            cfg = self.config["formatter"]

            # Before 0.6, configuration is a simple (unordered) dictionary:
            it: Iterator[FormatterSpec]
            if isinstance(cfg, dict):
                it = iter(cfg.items())
            # After 0.6, configuration is ordered list of dictionaries,
            # Convert to ordered list of tuples.
            else:
                it = (next(iter(f.items())) for f in cfg)

            return [bento.formatter.for_name(f_class, self, cfg) for f_class, cfg in it]



import logging
import subprocess
import sys
import time
from datetime import datetime
from functools import update_wrapper, wraps
from typing import Any, Callable, List, Optional

import click

import bento.metrics
import bento.network
from bento.error import BentoException
from bento.util import echo_error, echo_warning

_AnyCallable = Callable[..., Any]


def echo_exception(f: _AnyCallable) -> _AnyCallable:
    @wraps(f)
    def new_func(*args: Any, **kwargs: Any) -> Any:
        try:
            f(*args, **kwargs)
        except BentoException as e:
            if e.msg:
                echo_error(e.msg)
            sys.exit(3)

    return new_func


def __log_exception(e: Exception) -> None:
    logging.exception(e)
    if isinstance(e, subprocess.CalledProcessError):
        cmd = e.cmd
        if isinstance(e.cmd, list):
            cmd = " ".join([str(part) for part in e.cmd])
        echo_warning(f'Could not execute "{cmd}":\n{e.stderr}')
        logging.error(e.stdout)
        logging.error(e.stderr)
    else:
        echo_error(f"There was an exception {e}")


def with_metrics(f: _AnyCallable) -> _AnyCallable:
    def new_func(*args: Any, **kwargs: Any) -> Any:
        exit_code = 0
        before = time.time()

        context = click.get_current_context()

        commands: List[str] = [context.command.name]
        current_context = context
        while current_context.parent:
            current_context = current_context.parent

            # Don't include "cli" in the command string. This is a quirk in click command nesting
            if current_context.command.name != "cli":
                commands.insert(0, current_context.command.name)

        command_str: str = " ".join(commands)

        logging.info(f"command: {command_str}")

        cli_context = context.obj
        timestamp = (
            cli_context.timestamp if cli_context else datetime.utcnow().isoformat("T")
        )
        logging.info(f"Executing {command_str}")

        exception: Optional[BaseException] = None
        try:
            res = f(*args, **kwargs)
        except KeyboardInterrupt as e:
            # KeyboardInterrupt is a BaseException and has no exit code. Use 130 to mimic bash behavior.
            exit_code = 130
            exception = e
        except BentoException as e:
            exit_code = e.code
            exception = e
        except SystemExit as e:
            exit_code = e.code
            exception = e
        except Exception as e:
            exit_code = 3
            __log_exception(e)
            exception = e

        exc_name = exception.__class__.__name__ if exception else None

        elapsed = time.time() - before
        user_duration = cli_context.user_duration() if cli_context else None

        if exc_name == "KeyboardInterrupt":
            logging.info(f"{command_str} interrupted after running for {elapsed}s")
        else:
            logging.info(
                f"{command_str} completed in {elapsed}s with exit code {exit_code}"
            )

        email = (
            cli_context.email
            if cli_context and cli_context.email
            else bento.metrics.read_user_email()
        )

        logging.info(
            f"______________exit code: {exit_code}, exception name: {exc_name}_______________"
        )

        metrics = bento.metrics.command_metric(
            command_str,
            email,
            timestamp,
            kwargs,
            exit_code,
            elapsed,
            exc_name,
            user_duration,
        )
        if not cli_context.headless:
            bento.network.post_metrics(metrics)
        if exception:
            raise exception
        return res

    return update_wrapper(new_func, f)



import json
import logging
import subprocess
from abc import abstractmethod
from pathlib import Path
from typing import Dict, Optional, Set

import attr
from semantic_version import NpmSpec, Version

from bento.error import NodeError
from bento.tool.tool import Tool

NODE_VERSION_RANGE = NpmSpec("^8.10.0 || ^10.13.0 || >=11.10.1")

VersionDict = Dict[str, Version]


@attr.s(auto_attribs=True)
class NpmDeps:
    """Represents top-level npm package dependencies for a project"""

    main: Dict[str, NpmSpec]
    dev: Dict[str, NpmSpec]

    def __contains__(self, item: str) -> bool:
        return item in self.main or item in self.dev


class JsTool(Tool):
    @property
    @abstractmethod
    def install_location(self) -> Path:
        pass

    def _dependencies(self, location: Path) -> NpmDeps:
        """
        Returns an inventory of all top-level dependencies (not necessarily installed)

        :return: The dependencies, with versions
        """
        package_json_path = location / "package.json"
        with package_json_path.open() as stream:
            packages = json.load(stream)

        return NpmDeps(
            packages.get("dependencies", {}), packages.get("devDependencies", {})
        )

    def _installed_version(self, package: str, location: Path) -> Optional[Version]:
        """
        Gets the version of a package that was installed.

        Returns None if that package has not been installed.
        """
        package_json_path = location / "node_modules" / package / "package.json"
        try:
            with package_json_path.open() as f:
                package_json = json.load(f)
        except FileNotFoundError:
            return None
        if "version" in package_json:
            return Version(package_json["version"])
        else:
            return None

    def _npm_install(self, packages: VersionDict) -> None:
        """Runs npm install $package@^$version for each package."""
        logging.info(f"Installing {packages}...")
        args = [f"{name}@^{version}" for name, version in packages.items()]
        cmd = ["npm", "install", "--save-dev"]
        result = self.execute(
            cmd + args,
            cwd=self.install_location,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        logging.info(result.stdout)
        logging.info(result.stderr)

    def _ensure_packages(self, packages: VersionDict) -> Set[str]:
        """Ensures that the given packages are installed.

        Returns the list of all packages that were installed.

        The argument maps package names to the minimum version. This is morally
        equivalent to a plain `npm install --save`, except it's faster in the
        case where all the packages are already installed.
        """
        to_install = {}
        for name, required_version in packages.items():
            installed = self._installed_version(name, location=self.install_location)
            if not (installed and installed >= required_version):
                to_install[name] = required_version
        if not to_install:
            return set()

        self._npm_install(to_install)
        return set(to_install)

    def _ensure_node_version(self) -> None:
        """Ensures that Node.js version installed on the system is compatible with ESLint v6
        per https://github.com/eslint/eslint/blob/master/docs/user-guide/migrating-to-6.0.0.md#-nodejs-6-is-no-longer-supported
        Suppored Node.js version  ^8.10.0 || ^10.13.0 || >=11.10.1
        """
        version = self.execute(
            ["node", "--version"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        node_version = version.stdout.rstrip().strip("v")
        if version.returncode > 0 or not NODE_VERSION_RANGE.match(
            Version(node_version)
        ):
            raise NodeError(
                f"Node.js is not installed, or its version is not >8.10.0, >10.13.0, or >=11.10.1 (found {node_version})."
            )



import logging
from collections import namedtuple
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, List

import attr
import click
from pre_commit.git import zsplit
from pre_commit.staged_files_only import staged_files_only
from pre_commit.util import CalledProcessError, cmd_output, noop_context

import bento.base_context
import bento.git
from bento.error import UnsupportedGitStateException
from bento.fignore import FileIgnore, Parser
from bento.tool_runner import RunStep
from bento.util import Colors, echo_error, echo_newline

if TYPE_CHECKING:
    # Only import when type checking to avoid loading module when unecessary
    import git  # noqa

PATCH_CACHE = str(Path.home() / ".cache" / "bento" / "patches")
HEAD = "HEAD"

GitStatus = namedtuple("GitStatus", ["added", "modified", "removed", "unmerged"])


class StatusCode:
    Added = "A"
    Deleted = "D"
    Renamed = "R"
    Modified = "M"
    Unmerged = "U"
    Untracked = "?"
    Ignored = "!"


class NoGitHeadException(Exception):
    """
    Raised by TargetFileManager when trying to get a head context
    and no git head commit is found
    """

    pass


@attr.s
class TargetFileManager:
    """
        Handles all logic related to knowing what files to run on.

        This includes:
            - understanding files are ignores based on bentoignore rules
            - traversing project directories
            - pruning traversal
            - listing staged files
            - changing git state

        Parameters:
            base_path: Path to start walking files from
            paths: List of Paths (absolute or relative to current working directory) that
                    we want to traverse
            staged: whether we want to scan just staged files
            base_commit: Git ref to compare against in --staged mode (default HEAD)
            ignore_rules_file_path: Path to .bentoignore file
    """

    _base_path = attr.ib(type=Path)
    _paths = attr.ib(type=List[Path])
    _staged = attr.ib(type=bool)
    _ignore_rules_file_path = attr.ib(type=Path)
    _base_commit = attr.ib(type=str, default=HEAD)
    _status = attr.ib(type=GitStatus, init=False)
    _target_paths = attr.ib(type=List[Path], init=False)

    def _fname_to_path(self, repo: "git.Repo", fname: str) -> Path:
        return (Path(repo.working_tree_dir) / fname).resolve()

    @_status.default
    def get_git_status(self) -> GitStatus:
        """
            Returns Absolute Paths to all files that are staged
        """
        import gitdb.exc  # type: ignore

        repo = bento.git.repo()

        if not repo:
            return GitStatus([], [], [], [])

        try:
            repo.rev_parse(self._base_commit)
            has_commit = True
        except gitdb.exc.BadName:
            has_commit = False

        if not has_commit and self._base_commit != HEAD:
            raise UnsupportedGitStateException(
                f"Unknown git ref '{self._base_commit}'. To run against all diffs on this repository, please use --all."
            )

        # Output of git command will be relative to git project root
        cmd = [
            "git",
            "diff",
            "--name-status",
            "--no-ext-diff",
            "-z",
            "--diff-filter=ACDMRTUXB",
        ]
        if self._staged:
            cmd.append("--staged")
        if has_commit:
            cmd.append(self._base_commit)
        result = repo.git.execute(cmd)
        status_output = zsplit(result)

        added = []
        modified = []
        removed = []
        unmerged = []
        while status_output:
            code = status_output[0]
            fname = status_output[1]
            trim_size = 2

            if not code.strip():
                continue
            if code == StatusCode.Untracked or code == StatusCode.Ignored:
                continue

            # The following detection for unmerged codes comes from `man git-status`
            if code == StatusCode.Unmerged:
                unmerged.append(self._fname_to_path(repo, fname))
            if (
                code[0] == StatusCode.Renamed
            ):  # code is RXXX, where XXX is percent similarity
                removed.append(self._fname_to_path(repo, fname))
                fname = status_output[2]
                trim_size += 1
                added.append(self._fname_to_path(repo, fname))
            if code == StatusCode.Added:
                added.append(self._fname_to_path(repo, fname))
            if code == StatusCode.Modified:
                modified.append(self._fname_to_path(repo, fname))
            if code == StatusCode.Deleted:
                removed.append(self._fname_to_path(repo, fname))

            status_output = status_output[trim_size:]
        logging.info(
            f"Git status:\nadded: {added}\nmodified: {modified}\nremoved: {removed}\nunmerged: {unmerged}"
        )
        return GitStatus(added, modified, removed, unmerged)

    @_target_paths.default
    def _get_target_files(self) -> List[Path]:
        """
            Return list of all absolute paths to analyze
        """
        # resolve given paths relative to current working directory
        paths = [p.resolve() for p in self._paths]

        # If staged then only run on files that are different
        # and are a subpath of anything in input_paths
        if self._staged:
            paths = [
                a
                for a in (self._status.added + self._status.modified)
                # diff_path is a subpath of some element of input_paths
                if any((a == path or path in a.parents) for path in paths)
            ]

        # Filter out ignore rules, expand directories
        ignore_path = (
            self._ignore_rules_file_path
            if self._ignore_rules_file_path.exists()
            else bento.base_context.DEFAULT_IGNORE_PATH
        )
        with ignore_path.open() as ignore_lines:
            patterns = Parser(self._base_path, ignore_path).parse(ignore_lines)

        file_ignore = FileIgnore(
            base_path=self._base_path, patterns=patterns, target_paths=paths
        )

        filtered: List[Path] = []
        for elem in file_ignore.entries():
            if elem.survives:
                filtered.append(elem.path)

        return filtered

    def _abort_if_untracked_and_removed(
        self, working_path: Path, removed: List[Path]
    ) -> None:
        """
            Raises UnsupportedGitStateException if any path is removed from
            the git index but also appears in the filesystem.

            :param removed (list): Removed paths
            :raises UnsupportedGitStateException: If any removed paths are present on Filesystem
        """
        untracked_removed = [
            str(r.relative_to(working_path)).replace(" ", r"\ ")
            for r in removed
            if r.exists()
        ]
        if untracked_removed:
            joined = " ".join(untracked_removed)

            def echo_cmd(cmd: str) -> None:
                click.echo(f"    $ {click.style(cmd, bold=True)}\n", err=True)

            echo_error(
                "One or more files deleted from git exist on the filesystem. Aborting to prevent data loss. To "
                "continue, please stash by running the following two commands:"
            )
            echo_newline()
            echo_cmd(f"git stash -u -- {joined}")
            echo_cmd(f"git rm {joined}")
            click.secho(
                "Stashed changes can later be recovered by running:\n",
                err=True,
                fg=Colors.ERROR,
            )
            echo_cmd(f"git stash pop")
            raise UnsupportedGitStateException()

    @contextmanager
    def _head_context(self) -> Iterator[None]:
        """
        Runs a block of code on files from the current branch HEAD.

        :raises subprocess.CalledProcessError: If git encounters an exception
        :raises NoGitHeadException: If git cannot detect a HEAD commit
        :raises UnsupportedGitStateException: If unmerged files are detected
        """
        repo = bento.git.repo()

        if not repo:
            yield
            return

        commit = bento.git.commit()
        if commit is None:
            raise NoGitHeadException()

        else:
            # Need to look for unmerged files first, otherwise staged_files_only will eat them
            if self._status.unmerged:
                echo_error(
                    "Please resolve merge conflicts in these files before continuing:"
                )
                for f in self._status.unmerged:
                    click.secho(str(f.relative_to(repo.working_tree_dir)), err=True)
                raise UnsupportedGitStateException()

            with staged_files_only(PATCH_CACHE):
                tree = cmd_output("git", "write-tree")[1].strip()
                self._abort_if_untracked_and_removed(
                    repo.working_tree_dir, self._status.removed
                )
                try:
                    for a in self._status.added:
                        a.unlink()
                    cmd_output("git", "checkout", self._base_commit, "--", ".")
                    yield
                finally:
                    # git checkout will fail if the checked-out index deletes all files in the repo
                    # In this case, we still want to continue without error.
                    # Note that we have no good way of detecting this issue without inspecting the checkout output
                    # message, which means we are fragile with respect to git version here.
                    try:
                        cmd_output("git", "checkout", tree.strip(), "--", ".")
                    except CalledProcessError as ex:
                        if (
                            ex.output
                            and len(ex.output) >= 2
                            and "pathspec '.' did not match any file(s) known to git"
                            in ex.output[1].strip()
                        ):
                            logging.warning(
                                "Restoring git index failed due to total repository deletion; skipping checkout"
                            )
                        else:
                            raise ex
                    if self._status.removed:
                        cmd_output(
                            "git", "rm", "-f", *(str(r) for r in self._status.removed)
                        )

    @contextmanager
    def run_context(self, staged: bool, run_step: RunStep) -> Iterator[List[Path]]:
        """
        Provides a context within which to run tools and returns list of paths
        that should be analyzed.

        Possible contexts include:

            Head Context: all files in current branch HEAD
            Staged Files Context: all files in current branch HEAD plus staged changes
                                  (hides all untracked files)
            Noop: all files as currently available on filesystem

        Returned list of paths are all abolute paths and include all files that are
            - not ignored based on .bentoignore rules and
            - exist in any path filters specified.

        :param staged: Whether to use remove file diffs
        :param run_step: Which run step is in use (baseline if tool is determining baseline, check if tool is finding new results)
        :return: A Python with-expression
        :raises subprocess.CalledProcessError: If git encounters an exception
        :raises NoGitHeadException: If git cannot detect a HEAD commit
        :raises UnsupportedGitStateException: If unmerged files are detected
        """
        if staged and run_step == RunStep.BASELINE:
            stash_context = self._head_context()
        elif staged:
            stash_context = staged_files_only(PATCH_CACHE)
        else:
            # staged is False
            stash_context = noop_context()

        with stash_context:
            yield self._target_paths



import logging
import subprocess
import sys
import threading
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple

import click

import bento.constants
import bento.formatter
import bento.metrics
import bento.network
import bento.orchestrator
import bento.result
import bento.tool_runner
from bento.config import get_valid_tools, update_tool_run
from bento.context import Context
from bento.decorators import with_metrics
from bento.error import (
    BentoException,
    IncompatibleParametersException,
    NodeError,
    ToolRunException,
)
from bento.paths import list_paths
from bento.result import Baseline
from bento.target_file_manager import TargetFileManager
from bento.tool import Tool
from bento.util import echo_error, echo_next_step, echo_success, echo_warning
from bento.violation import Violation

OVERRUN_PAGES = 3


def __get_ignores_for_tool(tool: str, config: Dict[str, Any]) -> List[str]:
    tool_config = config["tools"]
    tool_specific_config = tool_config[tool] or {}
    return tool_specific_config.get("ignore", [])


def __headed_help(prog_name: str) -> str:
    return f"""
Checks for new findings.

By default, only staged files are checked. New findings introduced by
these staged changes AND that are not in the archive (`.bento/archive.json`)
will be shown.

Use `--all` to check all Git tracked files, not just those that are staged:

    $ {prog_name} check --all [PATHS]

Optional PATHS can be specified to check specific directories or files.

See `{prog_name} archive --help` to learn about suppressing findings.
"""


def __headless_help(prog_name: str) -> str:
    return f"""
Checks for new findings.

By default, only staged files are checked. Only new findings introduced by
these staged changes will be shown.

Use `--all` to check all Git tracked files, not just those that are staged:

    $ {prog_name} check --all [PATHS]

Optional PATHS can be specified to check specific directories or files.
"""


@click.command()
@click.option(
    "--all",
    "all_",
    is_flag=True,
    default=False,
    help="Show unarchived findings for all tracked files. Incompatible with --diff-against.",
)
@click.option(
    "-f",
    "--formatter",
    type=click.Choice(bento.formatter.FORMATTERS.keys()),
    help=f"Which output format to use. Falls back to the formatter(s) configured in `{bento.constants.CONFIG_FILE_NAME}`.",
    multiple=True,
)
@click.option(
    "--pager/--no-pager",
    help="Send long output through a pager. This should be disabled when used as an integration (e.g. with an editor).",
    default=True,
)
@click.option(
    "-t",
    "--tool",
    help="Specify a tool to run.",
    metavar="TOOL",
    multiple=True,
    default=[],
    autocompletion=get_valid_tools,
)
@click.option(
    "--diff-against",
    help="Git commit versus which to compare results. Defaults to 'HEAD'. Incompatible with --all.",
)
@click.option("--staged-only", is_flag=True, default=False, hidden=True)
@click.argument("paths", nargs=-1, type=Path, autocompletion=list_paths)
@click.pass_obj
@with_metrics
def check(
    context: Context,
    all_: bool = False,
    formatter: Tuple[str, ...] = (),
    pager: bool = True,
    tool: Optional[List[str]] = None,
    diff_against: Optional[str] = None,
    staged_only: bool = False,  # Should not be used. Legacy support for old pre-commit hooks
    paths: Tuple[Path, ...] = (),
) -> None:
    context.validate_init()

    # Default to no path filter
    path_list = list(paths) if len(paths) > 0 else [context.base_path]

    # Handle specified tool that is not configured
    for t in (t for t in tool or [] if t not in context.configured_tools):
        click.echo(
            f"{tool} has not been configured. Adding default configuration for tool to {bento.constants.CONFIG_FILE_NAME}"
        )
        update_tool_run(context, t, False)
        # Set configured_tools to None so that future calls will
        # update and include newly added tool
        context._configured_tools = None

    # Handle specified formatters
    if formatter:
        context.config["formatter"] = [{f: {}} for f in formatter]

    if all_:
        click.echo(f"Running checks on all tracked files...\n", err=True)
    else:
        click.echo(f"Running checks on staged files...\n", err=True)

    tools: Iterable[Tool[Any]] = context.tools.values()
    if not tool and context.headless:
        tools = [context.configured_tools["semgrep"]]
    if tool:
        tools = [context.configured_tools[t] for t in tool]

    baseline: Baseline = {}
    if context.baseline_file_path.exists():
        with context.baseline_file_path.open() as json_file:
            baseline = bento.result.json_to_violation_hashes(json_file)

    if diff_against and all_:
        raise IncompatibleParametersException(["--base", "--all"])
    base_commit = diff_against or "HEAD"

    target_file_manager = TargetFileManager(
        context.base_path, path_list, not all_, context.ignore_file_path, base_commit
    )

    all_results, elapsed = bento.orchestrator.orchestrate(
        baseline, target_file_manager, not all_, tools
    )

    fmts = context.formatters
    findings_to_log: List[Any] = []
    n_all = 0
    n_all_filtered = 0
    filtered_findings: Dict[str, List[Violation]] = {}
    for tool_id, findings in all_results:
        if isinstance(findings, Exception):
            logging.error(findings)
            echo_error(f"Error while running {tool_id}: {findings}")
            if isinstance(findings, BentoException):
                click.secho(findings.msg, err=True)
            else:
                if isinstance(findings, subprocess.CalledProcessError):
                    click.secho(findings.stderr, err=True)
                    click.secho(findings.stdout, err=True)
                if isinstance(findings, NodeError):
                    echo_warning(
                        f"Node.js not found or version is not compatible with ESLint v6."
                    )
                click.secho(
                    f"""
    You can view full details of this error in `{bento.constants.DEFAULT_LOG_PATH}`.
    """,
                    err=True,
                )
            context.error_on_exit(ToolRunException())
        elif isinstance(findings, list) and findings:
            findings_to_log += bento.metrics.violations_to_metrics(
                tool_id,
                context.timestamp,
                findings,
                __get_ignores_for_tool(tool_id, context.config),
            )
            filtered = [f for f in findings if not f.filtered]
            filtered_findings[tool_id] = filtered

            n_all += len(findings)
            n_filtered = len(filtered)
            n_all_filtered += n_filtered
            logging.debug(f"{tool_id}: {n_filtered} findings passed filter")

    def post_metrics() -> None:
        if not context.headless:
            bento.network.post_metrics(findings_to_log, is_finding=True)

    stats_thread = threading.Thread(name="stats", target=post_metrics)
    stats_thread.start()

    dumped = [f.dump(filtered_findings) for f in fmts]
    context.start_user_timer()
    bento.util.less(dumped, pager=pager, overrun_pages=OVERRUN_PAGES)
    context.stop_user_timer()

    finding_source_text = "in this project" if all_ else "due to staged changes"
    if n_all_filtered > 0:
        echo_warning(
            f"{n_all_filtered} finding(s) {finding_source_text} in {elapsed:.2f} s"
        )
        click.secho("\nPlease fix these issues, or:\n", err=True)
        echo_next_step(
            "To archive findings as tech debt", f"{context.prog_name} archive"
        )
        if not context.headless:
            echo_next_step(
                "To disable a specific check",
                f"{context.prog_name} disable check TOOL CHECK",
            )
    else:
        echo_success(f"0 findings {finding_source_text} in {elapsed:.2f} s\n")

    n_archived = n_all - n_all_filtered
    if n_archived > 0:
        echo_next_step(
            f"Not showing {n_archived} archived finding(s). To view",
            "cat .bento/archive.json",
        )

    if not all_ and not context.autorun_is_blocking:
        return
    elif context.on_exit_exception:
        raise context.on_exit_exception
    elif n_all_filtered > 0:
        sys.exit(2)


# TODO: make not monkey with globals
def make_check(prog_name: str, headless: bool) -> click.Command:
    check.help = __headed_help(prog_name) if headless else __headless_help(prog_name)
    return check



import logging
import os
import platform
import traceback
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from bento.util import EMPTY_DICT

BASE_URL = "https://bento.r2c.dev"
# Add default timeout so that we do not block the user's main thread from exiting
# 1 second value is so that user does not get impatient
TIMEOUT = 1  # sec

if TYPE_CHECKING:
    # Only import when type checking to avoid loading module when unecessary
    from requests.models import Response  # noqa


PostData = List[Dict[str, Any]]


def _get_version() -> str:
    from bento import __version__

    return __version__


def _get_default_shell() -> str:
    return os.environ.get("SHELL", "")


def _get_default_headers() -> Dict[str, str]:
    """
    Headers for all bento http/s requests
    """
    return {
        "X-R2C-BENTO-User-Platform": f"{platform.platform()}",
        "X-R2C-BENTO-User-Shell": f"{_get_default_shell()}",
        "X-R2C-BENTO-Cli-Version": f"{_get_version()}",
        "Accept": "application/json",
    }


def no_auth_get(
    url: str,
    params: Dict[str, str] = EMPTY_DICT,
    headers: Dict[str, str] = EMPTY_DICT,
    **kwargs: Any,
) -> "Response":
    # import inside def for performance
    import requests

    """Perform a requests.get and default headers set"""
    headers = {**_get_default_headers(), **headers}
    r = requests.get(
        url, headers=headers, params=params, **{"timeout": TIMEOUT, **kwargs}
    )
    return r


def no_auth_post(
    url: str,
    json: Any = EMPTY_DICT,
    params: Dict[str, str] = EMPTY_DICT,
    headers: Dict[str, str] = EMPTY_DICT,
    timeout: float = TIMEOUT,
) -> "Response":
    # import inside def for performance
    import requests

    """Perform a requests.post and default headers set"""
    headers = {**_get_default_headers(), **headers}
    r = requests.post(url, headers=headers, params=params, json=json, timeout=timeout)
    return r


def _get_base_url() -> str:
    return BASE_URL


def fetch_latest_version() -> Tuple[Optional[str], Optional[str]]:
    try:
        url = f"{_get_base_url()}/bento/api/v1/version"
        r = no_auth_get(url)
        response_json = r.json()
        return response_json.get("latest", None), response_json.get("uploadTime", None)
    except Exception as e:
        logging.exception(e)
        return None, None


def post_metrics(data: PostData, is_finding: bool = False) -> bool:
    try:
        url = f"{_get_base_url()}/bento/api/v4/metrics/"
        if is_finding:
            url = f"{url}finding/"
        # logging.debug(data)  # This spams the log when many findings
        r = no_auth_post(url, json=data)
        r.raise_for_status()
        return True
    except Exception as e:
        logging.warning(
            f"Exception while posting metrics {e}\n{traceback.format_exc()}"
        )
        return False



import json
import os
import re
from pathlib import Path
from typing import Any, Dict, Iterable, List, Pattern, Type

from bento.parser import Parser
from bento.tool import output
from bento.util import fetch_line_in_file
from bento.violation import Violation

# Input example:
# { "line": 49,
# "column": 49,
# "path": "tests/test_tool.py",
# "code": 11,
# "name": "Undefined type",
# "description": "Undefined type [11]: Type `Iterable` is not defined.",
# "long_description": "Undefined type [11]: Type `Iterable` is not defined.",
# "concise_description": "Undefined type [11]: Type `Iterable` is not defined.",
# "inference": {},
# "ignore_error": false,
# "external_to_global_root": false
# }

PINNED_PYRE_VERSION = "0.0.32"


class PyreParser(Parser):
    def to_violation(self, result: Dict[str, Any]) -> Violation:
        path = self.trim_base(result["path"])
        abspath = self.base_path / path

        check_id = str(result["code"])
        line = result["line"]
        return Violation(
            tool_id=PyreTool.TOOL_ID,
            check_id=check_id,
            path=path,
            line=line,
            column=result["column"],
            message=result["description"],
            severity=2,
            syntactic_context=fetch_line_in_file(abspath, line) or "<no source found>",
            link="https://pyre-check.org/docs/error-types.html",
        )

    def parse(self, tool_output: str) -> List[Violation]:
        results: List[Dict[str, Any]] = json.loads(tool_output)
        return [self.to_violation(r) for r in results]


class PyreTool(output.Str):
    TOOL_ID = "pyre"  # to-do: versioning?
    PROJECT_NAME = "Python"

    @property
    def parser_type(self) -> Type[Parser]:
        return PyreParser

    @classmethod
    def tool_id(self) -> str:
        return PyreTool.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Finds type errors in Python code"

    @property
    def project_name(self) -> str:
        return PyreTool.PROJECT_NAME

    @property
    def file_name_filter(self) -> Pattern:
        return re.compile(r".*\.py\b")

    def matches_project(self, files: Iterable[Path]) -> bool:
        # disabled by default for now
        return False

    def setup(self) -> None:
        cmd = ["pip3", "install", "-q", f"pyre-check=={PINNED_PYRE_VERSION}"]
        self.execute(cmd)

    def run(self, files: Iterable[str]) -> str:
        # for pyre, we also want to make sure that we are running in the same venv environment that the
        # source project is in, so that all the dependencies and types are resolvable
        if not os.environ.get("PIPENV_ACTIVE") == "1":
            raise Exception(
                "pyre requires being run from this project's pipenv, but no pipenv is active"
            )

        # TODO: we haven't verified that we're in the *right* pipenv, just that we're in *a* pipenv
        # TODO: may need to do this *in parallel* for multiple subprojects inside a monorepo

        ALL_SOURCE_DIRS = [f"--source-directory={source_dir}" for source_dir in files]
        cmd = (
            ["pyre", "--noninteractive", "--output=json"] + ALL_SOURCE_DIRS + ["check"]
        )
        return self.execute(cmd, capture_output=True).stdout



import os
from pathlib import Path
from typing import Generic, List, TypeVar

import attr

from bento.violation import Violation

R = TypeVar("R", contravariant=True)


def _absolute(path: Path) -> Path:
    return path.absolute()


@attr.s
class Parser(Generic[R]):
    base_path = attr.ib(type=Path, converter=_absolute)

    def trim_base(self, path: str) -> str:
        wrapped = Path(path)
        if not wrapped.is_absolute():
            wrapped = self.base_path / wrapped
        return os.path.relpath(wrapped, self.base_path)

    def parse(self, result: R) -> List[Violation]:
        return []



import logging
import os
import shutil
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator

import attr
import click
import yaml

import bento.constants as constants
import bento.content.init as content
import bento.git
import bento.tool_runner
from bento.commands.autorun import install_autorun
from bento.commands.ci import install_ci, is_ci_configured, is_ci_provider_supported
from bento.context import Context
from bento.decorators import with_metrics
from bento.error import NotAGitRepoException
from bento.target_file_manager import TargetFileManager


@contextmanager
def interaction(context: click.Context) -> Iterator[None]:
    context.obj.start_user_timer()
    yield
    context.obj.stop_user_timer()


@attr.s(auto_attribs=True)
class InitCommand:
    context: Context

    def _install_config_if_not_exists(self) -> bool:
        """
        Installs .bento.yml if one does not already exist

        :return: whether a config was installed
        """
        config_path = self.context.config_path
        pretty_path = self.context.pretty_path(config_path)
        if not config_path.exists():
            on_done = content.InstallConfig.install.echo(pretty_path)
            with (
                open(os.path.join(os.path.dirname(__file__), "../configs/default.yml"))
            ) as template:
                yml = yaml.safe_load(template)

            target_file_manager = TargetFileManager(
                self.context.base_path,
                [self.context.base_path],
                staged=False,
                ignore_rules_file_path=self.context.ignore_file_path,
            )

            for tid, tool in self.context.tool_inventory.items():
                if (
                    not tool(self.context).matches_project(
                        target_file_manager._target_paths
                    )
                    and tid in yml["tools"]
                ):
                    del yml["tools"][tid]
            logging.debug(
                f"Matching tools for this project: {', '.join(yml['tools'].keys())}"
            )
            if not yml["tools"]:
                logging.warning("No tools match this project")
            with config_path.open("w") as config_file:
                yaml.safe_dump(yml, stream=config_file)
            on_done()
            logging.info(f"Created {pretty_path}.")
            return True
        else:
            content.InstallConfig.install.echo(pretty_path, skip=True)
            return False

    def _install_ignore_if_not_exists(self) -> bool:
        """
        Installs .bentoignore if it does not already exist

        :return: whether .bentoignore was created
        """
        pretty_path = self.context.pretty_path(self.context.ignore_file_path)
        if not self.context.ignore_file_path.exists():
            on_done = content.InstallIgnore.install.echo(pretty_path)
            templates_path = Path(os.path.dirname(__file__)) / ".." / "configs"
            shutil.copy(
                templates_path / constants.IGNORE_FILE_NAME,
                self.context.ignore_file_path,
            )

            gitignore_added = False

            # If we're in a git repo with a .gitignore, add it to .bentoignore
            repo = bento.git.repo()
            if repo:
                path_to_gitignore = Path(repo.working_tree_dir) / ".gitignore"
                if path_to_gitignore.exists():
                    include_path = os.path.relpath(
                        path_to_gitignore, self.context.base_path
                    )
                    with self.context.ignore_file_path.open("a") as ignore_file:
                        ignore_file.write(f":include {include_path}\n")
                    gitignore_added = True

            # Otherwise, add the contents of configs/extra-ignore-patterns
            if not gitignore_added:
                with (templates_path / "extra-ignore-patterns").open() as extras:
                    with self.context.ignore_file_path.open("a") as ignore_file:
                        for e in extras:
                            ignore_file.write(e)

            on_done()
            logging.info(f"Created {pretty_path}.")
            return True
        else:
            content.InstallIgnore.install.echo(pretty_path, skip=True)
            return False

    def _configure_autorun(self, ctx: click.Context, is_first_run: bool) -> None:
        if is_first_run:
            on_done = content.InstallAutorun.install.echo()
            ctx.invoke(install_autorun, block=True)
            on_done()
        else:
            content.InstallAutorun.install.echo(skip=True)

    def _should_we_install_ci(self, ctx: click.Context) -> bool:
        """
        Returns boolean indicating whether CI should be installed.

        The answer is no if CI isn't supported,
        or if the user asks not to configure it.
        """
        if not is_ci_provider_supported(self.context.base_path):
            return False

        if is_ci_configured(self.context):
            return False

        if not sys.stdin.isatty() or not sys.stderr.isatty():
            return False

        content.InstallCI.pitch.echo()

        with interaction(ctx):
            is_confirmed = content.InstallCI.confirm.echo()

        content.InstallCI.after_confirm.echo()
        return is_confirmed

    def _configure_ci(self, ctx: click.Context, is_ci_requested: bool) -> bool:
        """
        Sets up CI integration if the user asked for it.

        Returns whether CI has been newly installed.
        """
        config_path = self.context.pretty_path(self.context.gh_actions_file_path)
        config_directory = config_path.parts[0]

        if not is_ci_requested or is_ci_configured(self.context):
            content.InstallCI.progress.echo(config_directory, skip=True)
            return False

        on_done = content.InstallCI.progress.echo(config_directory)
        ctx.invoke(install_ci)
        on_done()

        return True

    def _install_tools(self, clean: bool) -> None:
        """
        Ensures tools are installed

        :param clean: If true, forces tool reinstallation
        """
        content.InstallTools.install.echo()
        if clean:
            content.Clean.tools.echo()
            shutil.rmtree(constants.VENV_PATH, ignore_errors=True)
        runner = bento.tool_runner.Runner(paths=[], install_only=True, use_cache=False)
        tools = self.context.tools.values()
        runner.parallel_results(tools, {})

    def _identify_git(self) -> None:
        repo = bento.git.repo(self.context.base_path)
        if repo is None:
            raise NotAGitRepoException()

    def _identify_project(self) -> None:
        """Identifies this project"""
        tools = self.context.tools.values()
        project_names = sorted(list({t.project_name for t in tools}))
        logging.debug(f"Project names: {project_names}")
        if len(project_names) > 2:
            projects = f'{", ".join(project_names[:-1])}, and {project_names[-1]}'
        elif project_names:
            projects = " and ".join(project_names)
        else:
            content.Identify.failure.echo()
            return
        content.Identify.success.echo(projects)

    def _finish(self, *, is_ci_newly_installed: bool) -> None:
        content.Finish.body.echo()
        if is_ci_newly_installed:
            content.InstallCI.finalize_ci.echo()

    def run(self, ctx: click.Context, clean: bool) -> None:
        content.Start.banner.echo()

        is_ci_requested = self._should_we_install_ci(ctx)

        if sys.stdin.isatty() and sys.stderr.isatty():
            with interaction(ctx):
                content.Start.confirm.echo(default=True, show_default=False)

        self.context.resource_path.mkdir(exist_ok=True)
        is_first_init = not self.context.config_path.exists()
        # Perform git identification
        self._identify_git()

        # Perform configuration
        self._install_ignore_if_not_exists()
        self._install_config_if_not_exists()
        self._configure_autorun(ctx, is_first_init)
        is_ci_newly_installed = self._configure_ci(ctx, is_ci_requested)

        # Perform project identification
        self._identify_project()

        # Perform initial analysis
        self._install_tools(clean)

        # Message next steps
        self._finish(is_ci_newly_installed=is_ci_newly_installed)


@click.command()
@click.pass_context
@click.option(
    "--clean",
    help="Reinstalls tools using a clean installation.",
    is_flag=True,
    default=False,
)
@with_metrics
def init(ctx: click.Context, clean: bool) -> None:
    """
    Autodetects and installs tools.

    Run again after changing tool list in .bento.yml
    """
    context: Context = ctx.obj

    InitCommand(context).run(ctx, clean)



import fnmatch
import logging
import os
import re
import time
from pathlib import Path
from typing import Collection, Dict, Iterable, Iterator, List, Mapping, Set, TextIO

import attr

import bento.constants as constants
from bento.util import echo_warning

CONTROL_REGEX = re.compile(r"(?!<\\):")  # Matches unescaped colons
MULTI_CHAR_REGEX = re.compile(
    r"(?!<\\)\[.*(?!<\\)\]"
)  # Matches anything in unescaped brackets
COMMENT_START_REGEX = re.compile(r"(?P<ignore_pattern>.*?)(?:\s+|^)#.*")


@attr.s
class Entry:
    path = attr.ib(type=Path)
    survives = attr.ib(type=bool)


@attr.s(auto_attribs=True)
class WalkEntries(Collection[Entry]):
    cache: Dict[Path, Entry]

    def __len__(self) -> int:
        return len(self.cache)

    def __iter__(self) -> Iterator[Entry]:
        return iter(self.cache.values())

    def __contains__(self, item: object) -> bool:
        return isinstance(item, Entry) and item.path in self.cache


@attr.s
class FileIgnore(Mapping[Path, Entry]):
    base_path = attr.ib(type=Path)
    patterns = attr.ib(type=Set[str])
    target_paths = attr.ib(type=List[Path])
    _processed_patterns = attr.ib(type=Set[str], init=False)
    _walk_cache: Dict[Path, Entry] = attr.ib(default=None, init=False)

    def __attrs_post_init__(self) -> None:
        self._processed_patterns = Processor(self.base_path).process(self.patterns)
        self._init_cache()

    def _survives(self, path: Path) -> bool:
        """
        Determines if a single Path survives the ignore filter.
        """
        for p in self._processed_patterns:
            if path.is_dir() and p.endswith("/") and fnmatch.fnmatch(str(path), p[:-1]):
                return False
            if fnmatch.fnmatch(str(path), p):
                return False

            # Check any subpath of path satisfies a pattern
            # i.e. a/b/c/d is ignored with rule a/b
            # This is a hack. TargetFileManager should be pruning while searching
            # instead of post filtering to avoid this
            # Note: Use relative to base to avoid ignore rules firing on parent directories
            # i.e. /private/var/.../instabot should not be ignored with var/ rule
            # in instabot dir as base_path
            # Note: Append "/" to path before running fnmatch so **/pattern matches with pattern/stuff
            if p.endswith("/") and fnmatch.fnmatch(
                "/" + str(path.relative_to(self.base_path)), p + "*"
            ):
                return False
            if (
                p.endswith("/")
                and p.startswith(str(self.base_path))
                and fnmatch.fnmatch(str(path), p + "*")
            ):
                return False

        return True

    def _walk(self, this_path: str, root_path: str) -> Iterator[Entry]:
        """
        Walks path, returning an Entry iterator for each item.

        If an item is not ignored, it is traversed recursively. Traversal stops on
        ignored items.

        Recalculates on every call.
        """
        # Handle non existent paths passed to cli.
        # TODO handle further up
        if not Path(this_path).exists():
            return

        if Path(this_path).is_file():
            yield Entry(Path(this_path), self._survives(Path(this_path)))
        else:
            for e in os.scandir(this_path):
                if e.is_symlink():
                    continue
                elif self._survives(Path(e.path)):
                    before = time.time()
                    for ee in self._walk(e.path, root_path):
                        yield ee
                    filename = Path(this_path) / e.name
                    logging.debug(f"Scanned {filename} in {time.time() - before} s")
                else:
                    # TODO I think we can remove the false ones and have existence be survival
                    yield Entry(Path(e.path), False)

    def _init_cache(self) -> None:
        pretty_patterns = "\n".join(self.patterns)
        logging.info(f"Ignored patterns are:\n{pretty_patterns}")
        before = time.time()
        self._walk_cache = {}
        for target in self.target_paths:
            entries = self._walk(str(target), str(self.base_path))
            self._walk_cache.update(dict((e.path, e) for e in entries))
        logging.info(f"Loaded file ignore cache in {time.time() - before} s.")

    def entries(self) -> Collection[Entry]:
        """
        Returns all files that are not ignored, relative to the base path.
        """
        return WalkEntries(self._walk_cache)

    def filter_paths(self, paths: Iterable[Path]) -> List[Path]:
        abspaths = (p.absolute() for p in paths if p.exists())
        return [
            p
            for p in abspaths
            if p in self and self[p].survives or p.samefile(self.base_path)
        ]

    def __getitem__(self, item: Path) -> Entry:
        return self._walk_cache[item]

    def __iter__(self) -> Iterator[Path]:
        return iter(self._walk_cache)

    def __len__(self) -> int:
        return len(self._walk_cache)

    def __contains__(self, item: object) -> bool:
        return item in self._walk_cache


@attr.s(auto_attribs=True)
class Parser:
    r"""
    A parser for bentoignore syntax.

    bentoignore syntax mirrors gitignore syntax, with the following modifications:
    - "Include" patterns (lines starting with "!") are not supported.
    - "Character range" patterns (lines including a collection of characters inside brackets) are not supported.
    - An ":include ..." directive is added, which allows another file to be included in the ignore pattern list;
      typically this included file would be the project .gitignore. No attempt at cycle detection is made.
    - Any line beginning with a colon, but not ":include ", will raise a ValueError.
    - "\:" is added to escape leading colons.

    Unsupported patterns are silently removed from the pattern list (this is done so that gitignore files may be
    included without raising errors), although the removal will be logged.

    Unfortunately there's no available parser for gitignore syntax in python, so we have
    to make our own. The syntax is simple enough that we can just roll our own parser, so
    I deliberately skip using a parser generator or combinator library, which would either need to
    parse on a character-by-character basis or make use of a large number of regex scans.

    The parser steps are, for each line in the input stream:
    1. Remove comments
    2. Remove unsupported gitignore syntax
    3. Expand directives

    The end result of this parsing is a set of human-readable patterns corresponding to gitignore syntax.
    To use these patterns with fnmatch, however, a final postprocessing step is needed, achieved by calling
    Processor.process().

    :param base_path:   The path relative to which :include directives should be evaluated
    :param ignore_path: The path of the file being parsed (for logging only)
    """

    # Parser steps are each represented as Generators. This allows us to chain
    # steps, whether the step is a transformation, a filter, an expansion, or any combination thereof.

    base_path: Path
    # TODO remove this
    ignore_path: Path

    @staticmethod
    def remove_comments(line: str) -> Iterator[str]:
        """If a line has a comment, remove the comment and just return the ignore pattern
        """
        m = COMMENT_START_REGEX.match(line)
        if m:
            yield m.groupdict().get(
                "ignore_pattern", ""
            )  # return empty string if entire line is a comment
        else:
            yield line.rstrip()

    @staticmethod
    def filter_supported(line: str) -> Iterator[str]:
        """Remove unsupported gitignore patterns"""
        if not line:
            pass
        elif line.startswith("!") or MULTI_CHAR_REGEX.search(line):
            logging.warning(f"Skipping unsupported gitignore pattern '{line}'")
        else:
            yield line

    def expand_directives(self, line: str) -> Iterable[str]:
        """Load :include files"""
        if line.startswith(":include "):
            include_path = self.base_path / line[9:]
            with include_path.open() as include_lines:
                sub_base = include_path.parent.resolve()
                sub_parser = Parser(sub_base, include_path)
                return sub_parser.parse(include_lines)
        elif CONTROL_REGEX.match(line):
            raise ValueError(
                f"Unknown ignore directive in {self.ignore_path}: '{line}'"
            )
        else:
            return (line for _ in range(1))

    def parse(self, stream: TextIO) -> Set[str]:
        """Performs parsing of an input stream"""
        return {
            pattern
            for line in stream
            for no_comments in self.remove_comments(line)
            for supported in self.filter_supported(no_comments)
            for pattern in self.expand_directives(supported)
        }


@attr.s(auto_attribs=True)
class Processor:
    """
    A post-processor for parsed bentoignore files.

    The postprocessor is responsible for converting the parser's intermediate representation to a set of
    patterns compatible with fnmatch. The steps are:
    1. Unescape escape characters
    2. Convert gitignore patterns into fnmatch patterns
    """

    # Per Parser, each Processor step is represented as a Generator.

    base_path: Path

    @staticmethod
    def unescape(line: str) -> Iterator[str]:
        """Expands escape characters"""
        out = ""
        is_escape = False
        for c in line:
            if is_escape:
                out += c
                is_escape = False
            elif c == "\\":
                is_escape = True
            else:
                out += c
        yield out

    def to_fnmatch(self, pat: str) -> Iterator[str]:
        """Convert a single pattern from gitignore to fnmatch syntax"""
        if pat.rstrip("/").find("/") < 0:
            # Handles:
            #   file
            #   path/
            pat = os.path.join("**", pat)
        if pat.startswith("./") or pat.startswith("/"):
            # Handles:
            #   /relative/to/root
            #   ./relative/to/root
            pat = pat.lstrip(".").lstrip("/")
        if not pat.startswith("**"):
            # Handles:
            #   path/to/absolute
            #   */to/absolute
            #   path/**/absolute
            pat = os.path.join(self.base_path, pat)
        yield pat

    def process(self, pre: Iterable[str]) -> Set[str]:
        """Post-processes an intermediate representation"""
        return {
            pattern
            for pat in pre
            for unescaped in self.unescape(pat)
            for pattern in self.to_fnmatch(unescaped)
        }


def open_ignores(
    base_path: Path, ignore_path: Path, is_init: bool = False
) -> FileIgnore:
    """
    Opens this project's ignore file
    """
    if not ignore_path.exists():
        if not is_init:
            echo_warning(
                f"""'{os.path.relpath(ignore_path, os.getcwd())}' not found; using default ignore patterns.
Please run 'bento init' to configure a {constants.IGNORE_FILE_NAME} for your project.
"""
            )
        ignore_path = (
            Path(os.path.dirname(__file__)) / "configs" / constants.IGNORE_FILE_NAME
        )

    logging.info(f"Loading bento file ignores from {os.path.abspath(ignore_path)}")

    with ignore_path.open() as ignore_lines:
        patterns = Parser(base_path, ignore_path).parse(ignore_lines)
        return FileIgnore(
            base_path=base_path, patterns=patterns, target_paths=[base_path]
        )



#!/usr/bin/env python3

__name__ = "bento-headless"
__author__ = "Return To Corporation"
R2C_SUPPORT_EMAIL = "support@r2c.dev"

try:
    import pkg_resources

    __version__ = pkg_resources.get_distribution(__name__).version
except:  # noqa: E722
    __version__ = "0.0.0+unknown"



import os
import shutil
from pathlib import Path, PurePath
from typing import List, Optional, Type

import bento.constants as constants
from bento.extra.base_sgrep import BaseSgrepParser, BaseSgrepTool
from bento.parser import Parser


class SgrepParser(BaseSgrepParser):
    @classmethod
    def tool_id(cls) -> str:
        return SgrepTool.TOOL_ID


class SgrepTool(BaseSgrepTool):
    CONFIG_PATH = PurePath("semgrep.yml")
    CONFIG_ENV: str = "BENTO_REGISTRY"
    REGISTRY_ROOT: str = "https://sgrep.live/c/"
    PARSER = SgrepParser
    TOOL_ID = "semgrep"

    @classmethod
    def tool_id(cls) -> str:
        return cls.TOOL_ID

    def can_use_cache(self) -> bool:
        return self.CONFIG_ENV not in os.environ

    def get_config_path(self) -> Optional[Path]:
        if self.CONFIG_ENV not in os.environ:
            return self.context.resource_path / self.CONFIG_PATH
        else:
            return None

    def extra_cache_paths(self) -> List[Path]:
        cp = self.get_config_path()
        return [cp.resolve()] if cp else []

    @property
    def config_str(self) -> str:
        """
        Returns the configuration argument and optional path to pass to sgrep


        If the config environment variable is set, uses that value verbatim.

        Otherwise, ensures that the default config file exists and uses that location.
        """
        config_path = self.get_config_path()
        if config_path is None:
            return self.REGISTRY_ROOT + os.environ[self.CONFIG_ENV]

        if not config_path.exists():
            os.makedirs(self.base_path / constants.RESOURCE_PATH, exist_ok=True)
            template = (
                Path(os.path.dirname(__file__)).parent / "configs" / self.CONFIG_PATH
            )
            shutil.copy(template, config_path)
        return str(constants.RESOURCE_PATH / self.CONFIG_PATH)

    @property
    def parser_type(self) -> Type[Parser]:
        return self.PARSER



import os
import shutil
import stat
import sys
from pathlib import Path

import click

import bento.constants as constants
import bento.git
import bento.tool_runner
from bento.context import Context
from bento.decorators import with_metrics
from bento.error import ExistingGitHookException, NotAGitRepoException
from bento.util import echo_next_step, echo_success, echo_warning


def _is_bento_precommit(filename: Path) -> bool:
    if not filename.exists():
        return False
    with filename.open() as lines:
        return any(constants.BENTO_TEMPLATE_HASH in l for l in lines)


def _configure_block(context: Context, block: bool) -> None:
    config = context.config
    autorun_config = config.get(constants.AUTORUN, {})
    autorun_config[constants.AUTORUN_BLOCK] = block
    config[constants.AUTORUN] = autorun_config
    context.config = config


def _notify_install(context: Context, block: bool) -> None:
    if not context.is_init:
        block_prefix = "" if block else "non-"
        echo_success(f"Installed Bento autorun in {block_prefix}blocking mode.")
        echo_next_step("To uninstall Bento autorun", "bento disable autorun")

        desc_prefix, cmd_prefix = ("", "") if not block else ("non-", "no-")
        echo_next_step(
            f"To make autorun {desc_prefix}blocking",
            f"bento enable autorun --{cmd_prefix}block",
        )


@click.command(name="autorun")
@click.option(
    "--block/--no-block",
    default=True,
    help="If --block, commits will fail if Bento finds an issue.",
)
@click.pass_obj
@with_metrics
def install_autorun(context: Context, block: bool) -> None:
    """
    Configures Bento to automatically run on commits.

    Autorun is configured only for you; it does not affect
    other contributors to this project.

    Autorun is only configured for the project from which this
    command is run.

    By default, Bento will block commits if it finds an issue.
    To prevent autorun from blocking commits, run:

        $ bento enable autorun --no-block

    """
    import git  # import inside def for performance

    # Get hook path
    repo = bento.git.repo(context.base_path)
    if repo is None:
        raise NotAGitRepoException()

    _configure_block(context, block)

    hook_path = Path(git.index.fun.hook_path("pre-commit", repo.git_dir))

    if _is_bento_precommit(hook_path):
        _notify_install(context, block)

    else:
        legacy_hook_path = Path(f"{hook_path}.pre-bento")
        if hook_path.exists():
            # If pre-commit hook already exists move it over
            if legacy_hook_path.exists():
                raise ExistingGitHookException(str(hook_path))
            else:
                # Check that
                shutil.move(str(hook_path), str(legacy_hook_path))

        # Ensure .git/hooks directory exists
        # note that we can (and should) assume .git dir exists since
        # project must be a git project at this point in the code
        hook_path.parent.mkdir(exist_ok=True)

        # Copy pre-commit script template to hook_path
        template_location = os.path.join(
            os.path.dirname(__file__), "../configs/pre-commit.template"
        )
        shutil.copyfile(template_location, hook_path)

        # Make file executable
        original_mode = hook_path.stat().st_mode
        os.chmod(hook_path, original_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

        _notify_install(context, block)


@click.command(name="autorun")
@click.pass_obj
@with_metrics
def uninstall_autorun(context: Context) -> None:
    """
    Configures Bento to NOT run automatically on commits.

    Autorun is only removed for the project from which this
    command is run.
    """
    import git  # import inside def for performance

    # Get hook path
    repo = bento.git.repo(context.base_path)
    if repo is None:
        raise NotAGitRepoException()

    hook_path = Path(git.index.fun.hook_path("pre-commit", repo.git_dir))

    if not _is_bento_precommit(hook_path):
        echo_warning(
            "Not uninstalling autorun: Bento is not configured for autorun on this project."
        )
        sys.exit(1)
    else:
        # Put back legacy hook if exits
        legacy_hook_path = Path(f"{hook_path}.pre-bento")
        if legacy_hook_path.exists():
            shutil.move(str(legacy_hook_path), str(hook_path))
        else:
            hook_path.unlink()

        echo_success("Uninstalled Bento autorun.")
        echo_next_step("To enable autorun", "bento enable autorun")



from bento.renderer import (
    Confirm,
    Echo,
    Link,
    Multi,
    Newline,
    Processors,
    Progress,
    Steps,
    Sub,
    Text,
    Warn,
)


class Install:
    learn_more = Text(
        "To learn more, see the README.",
        processor=Processors.wrap_link(
            [
                Link(
                    "README",
                    "https://github.com/returntocorp/bento/blob/master/README.md#running-in-cicd",
                )
            ]
        ),
    )

    banner = Steps(
        Echo(
            Multi(
                [
                    "Configuring a GitHub Action to automatically check your team's pull requests. ",
                    learn_more,
                ],
                processor=Processors.wrap(),
            )
        ),
        Newline(),
    )

    progress = Progress(
        content=Multi(
            [
                "Creating GitHub Action configuration at ",
                Sub(0, style={"bold": True, "dim": True}),
                Text("/", style={"bold": True, "dim": True}),
            ]
        ),
        extra=24,
    )
    after_progress = Newline()

    finalize_ci = Steps(
        Warn("To finalize configuration of Bento's GitHub Action, please:"),
        Echo(
            Multi(
                [
                    "  $ git add .github/\n",
                    '  $ git commit -m "Add Bento GitHub Action"\n',
                    "  $ git push\n",
                ]
            )
        ),
    )


class Overwrite:
    warn = Steps(
        Echo(
            Multi(
                [
                    "It seems Bento is already installed for GitHub Actions at ",
                    Sub(0, style={"bold": True}),
                ]
            )
        ),
        Newline(),
        Warn(
            "If you continue, Bento will revert any changes "
            "you might have made manually to this CI configuration file."
        ),
        Newline(),
    )

    confirm = Confirm(
        "Do you want to revert Bento's GitHub Action to the default configuration?"
    )

    after_confirm = Newline()
